from typing import List

from .eng_tokenizer import EngTokenizer


class ZhCharTokenizer(EngTokenizer):

    """Chinese Character level tokenizer

    All of following functions are copied from BERT.
    https://github.com/google-research/bert/blob/master/tokenization.py

    E.g.
    >>> from uttut.pipeline.ops.zh_char_tokenizer import ZhCharTokenizer
    >>> op = ZhCharTokenizer()
    >>> output_seq, label_aligner = op.transform("這是a b")
    >>> output_labels = label_aligner.transform([1, 2, 3, 4, 5])
    >>> output_seq
    ["這", "是", "a", "b"]
    >>> output_labels
    [1, 2, 3, 5]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 2, 3, 0, 5]

    """

    def _tokenize(self, input_str: str) -> List[str]:
        text = self._tokenize_chinese_chars(input_str)
        return super()._tokenize(text)

    def _tokenize_chinese_chars(self, text):
        """Recognize CJK character and add whitespace around it

        E.g.
        1. "我喜歡吃蘋果" -> " 我  喜  歡  吃  蘋  果 "
        2. "我喜歡吃apples." -> " 我  喜  歡  吃 apples."

        """
        output = []
        for char in text:
            code_position = ord(char)
            if self._is_chinese_char(code_position):
                output.append(" ")
                output.append(char)
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _is_chinese_char(self, code_position):
        """Checks whether CP is the codepoint of a CJK character.

        This defines a "chinese character" as anything in the CJK Unicode block:
        https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)

        Note that the CJK Unicode block is NOT all Japanese and Korean characters,
        despite its name. The modern Korean Hangul alphabet is a different block,
        as is Japanese Hiragana and Katakana. Those alphabets are used to write
        space-separated words, so they are not treated specially and handled
        like the all of the other languages.

        """
        if ((code_position >= 0x4E00 and code_position <= 0x9FFF) or (
                code_position >= 0x3400 and code_position <= 0x4DBF) or (
                code_position >= 0x20000 and code_position <= 0x2A6DF) or (
                code_position >= 0x2A700 and code_position <= 0x2B73F) or (
                code_position >= 0x2B740 and code_position <= 0x2B81F) or (
                code_position >= 0x2B820 and code_position <= 0x2CEAF) or(
                code_position >= 0xF900 and code_position <= 0xFAFF) or (
                code_position >= 0x2F800 and code_position <= 0x2FA1F)):  #
            return True
        return False



from typing import List, Sequence, MutableSequence, Union

from .replacement cimport Replacement, ReplacementGroup  # noqa: E999


def _transform_sequence(
        input_seq: Union[str, List[str]],
        ReplacementGroup replacement_group,
        list output,
    ) -> Sequence:
    '''
    input_seq: list of str or pure str
    '''

    cdef unsigned int start, i
    cdef Replacement replacement

    start = 0
    i = 0
    for replacement in replacement_group:
        output[i] = input_seq[start: replacement.start]
        output[i + 1] = replacement.new_value
        start = replacement.end
        i += 2

    # tail
    output[-1] = input_seq[start:]

    return output


def _gen_inverse_replacement_group(
        input_seq: Union[str, List[str]],
        ReplacementGroup replacement_group,
    ) -> ReplacementGroup:
    '''
    input_seq: list of str or pure str
    '''

    cdef unsigned int n_replacement, start, dist
    cdef list dists
    cdef Replacement replacement
    cdef ReplacementGroup inverse_replacement_group

    inverse_replacement_group = ReplacementGroup()

    if replacement_group.is_empty():
        inverse_replacement_group.done()
        return inverse_replacement_group

    n_replacement = len(replacement_group)

    dists = get_dist_bt_replacement_group(replacement_group)
    dists.append(0)

    start = replacement_group[0].start

    for i in range(n_replacement):
        replacement = replacement_group[i]
        dist = dists[i]
        inverse_replacement_group.add(
            start=start,
            end=start + len(replacement.new_value),
            new_value=input_seq[replacement.start: replacement.end],
        )
        start += len(replacement.new_value) + dist
    inverse_replacement_group.done()
    return inverse_replacement_group


def get_dist_bt_replacement_group(ReplacementGroup replacement_group) -> List[int]:

    '''Compute the distance between replacement_group

    The distance is the length of sequence.

    Eg.
        Given [Replacement(0, 1, ''), Replacement(4, 7, '')]: [3]  # 4 - 1
        Given [Replacement(0, 3, ''), Replacement(3, 4, ''), Replacement(8, 11, '')]: [0, 4]
    Note that the length of output would be 1 less than that of input replacement_group.

    Arg:
        replacement_group (ReplacementGroup)

    Return:
        dist (ints)
    '''

    cdef unsigned int n_replacement, dist
    cdef list dists
    cdef Replacement current_replacement, next_replacement

    n_replacement = len(replacement_group)
    dists = [0] * (n_replacement - 1)

    for i in range(n_replacement - 1):
        current_replacement = replacement_group[i]
        next_replacement = replacement_group[i + 1]
        dist = next_replacement.start - current_replacement.end
        dists[i] = dist

    return dists



from typing import List

from ..label_transducer import get_most_common_except_not_entity
from ..utils.trie import Trie
from .base import Tokenizer, TokenizerAligner


class CustomWordTokenizerAligner(TokenizerAligner):

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)


class CustomWordTokenizer(Tokenizer):

    """Word Tokenizer Given User Words

    E.g.
    >>> from uttut.pipeline.ops.custom_word_tokenizer import CustomWordTokenizer
    >>> op = CustomWordTokenizer(['珍奶', '珍奶去冰', '去冰'])
    >>> output_seq, label_aligner = op.transform("一杯珍奶去冰")
    >>> output_labels = label_aligner.transform([1, 2, 3, 3, 3, 0])
    >>> output_seq
    ["一", "杯", "珍奶去冰"]
    >>> output_labels
    [1, 2, 3]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 2, 3, 3, 3, 3]

    """

    _label_aligner_class = CustomWordTokenizerAligner

    def __init__(self, user_words: List[str], shortest: bool = False):
        self.validate_user_words(user_words)
        self._user_words = user_words
        self._trie = Trie()
        for word in self._user_words:
            self._trie.insert(word)

        self._shortest = shortest

    @staticmethod
    def validate_user_words(user_words):
        if user_words is None or len(user_words) < 1:
            raise ValueError('User words should not be empty.')

    def _tokenize(self, input_str: str) -> List[str]:
        start = 0
        tokens = []
        while start < len(input_str):
            match_result = self._trie.match_prefix(
                input_str[start:],
                shortest=self._shortest,
            )
            if match_result is None:
                token = input_str[start: start + 1]
                start += 1
            else:
                token = match_result
                start += len(match_result)
            tokens.append(token)
        return tokens



from typing import List, Tuple

from .base import Operator, LabelAligner
from .tokens import START_TOKEN, END_TOKEN
from .label_transducer import get_not_entity
from ..edit import lst2lst
from ..edit.replacement import ReplacementGroup
from ..edit.label_propagation import propagate_by_replacement_group


class AddSosEos(Operator):

    """
    Add start token (<sos>) and end token (<eos>) to a sequence.

    E.g.
    >>> from uttut.pipeline.ops.add_sos_eos import AddSosEos
    >>> op = AddSosEos()
    >>> output_seq, label_aligner = op.transform(
        ['I', 'have', '10.7', 'dollars', '.'])
    >>> output_labels = label_aligner.transform([1, 2, 3, 4, 5])
    >>> output_seq
    ['<sos>', 'I', 'have', '10.7', 'dollars', '.', '<eos>']
    >>> output_labels
    [0, 1, 2, 3, 4, 5, 0]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 2, 3, 4, 5]

    """

    _input_type = list
    _output_type = list

    def __init__(
            self,
            start_token: str = START_TOKEN,
            end_token: str = END_TOKEN,
        ):
        self.start_token = start_token
        self.end_token = end_token

    def _transform(self, input_sequence: List[str]) -> Tuple[List[str], 'LabelAligner']:
        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = lst2lst.apply(input_sequence, forward_replacement_group)

        label_aligner = AddSosEosAligner(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )
        return output_sequence, label_aligner

    def _gen_forward_replacement_group(
            self,
            input_lst: List[str],
            annotation: str = None,
        ) -> ReplacementGroup:

        seqlen = len(input_lst)
        replacement_group = ReplacementGroup.add_all(
            [
                (0, 0, [self.start_token]),  # insert start token
                (seqlen, seqlen, [self.end_token]),  # insert end token
            ],
        )
        return replacement_group


class AddSosEosAligner(LabelAligner):

    def _transform(self, labels: List[int]) -> List[int]:
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=self._forward_edit,
            transduce_func=self._forward_transduce_func,
        )

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_not_entity(labels=labels, output_size=output_size)

    def _inverse_transform(self, labels: List[int]) -> List[int]:
        inverse_replacement_group = lst2lst.inverse(self._input_sequence, self._forward_edit)
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=inverse_replacement_group,
            transduce_func=self._backward_transduce_func,
        )

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_not_entity(labels=labels, output_size=output_size)



cdef extern from "MurmurHash3.h":  # noqa: E999
    void MurmurHash3_x86_32(
        void *key,  # noqa: E225
        int len,
        unsigned int seed,
        void *out,  # noqa: E225
    )


cdef unsigned int consistent_hash_in_c(  # noqa: E999
        str input_str,
        unsigned int mod_int,
        unsigned int seed,
    )



import random
from typing import List, Tuple, Union
from functools import reduce
from operator import mul

from .elements import (Datum, Entity)
from .toolkits.get_kth_combination import get_kth_combination
from .toolkits.partition_by_entities import partition_by_entities


__all__ = [
    'expand_by_entities',
]


def _aggregate_entities(
        list segments,  # noqa: E999
        list entity_labels,
    ):
    cdef list entities = []
    cdef int pointer = 0
    cdef int len_segment

    for seg, entity_label in zip(segments, entity_labels):
        len_seg = len(seg)
        if entity_label is None:
            pointer += len_seg
            continue

        entity = Entity(
            label=entity_label,
            value=seg,
            start=pointer,
            end=pointer + len_seg
        )
        entities.append(entity)
        pointer += len_seg
    return entities


def expand_by_entities(
        datum,
        sampling_method=None,
        bint include_orig=False,
    ):
    if not datum.has_entities():
        return [datum]

    if sampling_method is None:
        sampling_method = lambda n_combinations: list(range(n_combinations))  # noqa: E731
        # return all possible combinations

    parts, entity_labels = partition_by_entities(datum, include_orig)

    n_combinations = reduce(mul, [len(part) for part in parts])

    ints = sampling_method(n_combinations)

    result = []
    for idx in ints:
        segments = get_kth_combination(parts, idx)

        utterance = ''.join(segments)
        entities = _aggregate_entities(segments, entity_labels)
        new_datum = Datum(
            utterance=utterance,
            intents=datum.copy_intents(),
            entities=entities,
        )
        result.append(new_datum)
    return result



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tokenization classes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import re
import unicodedata
import six
# import tensorflow as tf


def validate_case_matches_checkpoint(do_lower_case, init_checkpoint):
    """Checks whether the casing config is consistent with the checkpoint name."""

    # The casing has to be passed in by the user and there is no explicit check
    # as to whether it matches the checkpoint. The casing information probably
    # should have been stored in the bert_config.json file, but it's not, so
    # we have to heuristically detect it to validate.

    if not init_checkpoint:
        return

    m = re.match("^.*?([A-Za-z0-9_-]+)/bert_model.ckpt", init_checkpoint)
    if m is None:
        return

    model_name = m.group(1)

    lower_models = [
        "uncased_L-24_H-1024_A-16", "uncased_L-12_H-768_A-12",
        "multilingual_L-12_H-768_A-12", "chinese_L-12_H-768_A-12",
    ]

    cased_models = [
        "cased_L-12_H-768_A-12", "cased_L-24_H-1024_A-16",
        "multi_cased_L-12_H-768_A-12",
    ]

    is_bad_config = False
    if model_name in lower_models and not do_lower_case:
        is_bad_config = True
        actual_flag = "False"
        case_name = "lowercased"
        opposite_flag = "True"

    if model_name in cased_models and do_lower_case:
        is_bad_config = True
        actual_flag = "True"
        case_name = "cased"
        opposite_flag = "False"

    if is_bad_config:
        raise ValueError(
            "You passed in `--do_lower_case=%s` with `--init_checkpoint=%s`. "
            "However, `%s` seems to be a %s model, so you "
            "should pass in `--do_lower_case=%s` so that the fine-tuning matches "
            "how the model was pre-training. If this error is wrong, please "
            "just comment out this check." % (actual_flag, init_checkpoint,
                                              model_name, case_name, opposite_flag))


def convert_to_unicode(text):
    """Converts `text` to Unicode (if it's not already), assuming utf-8 input."""
    if six.PY3:
        if isinstance(text, str):
            return text
        elif isinstance(text, bytes):
            return text.decode("utf-8", "ignore")
        else:
            raise ValueError("Unsupported string type: %s" % (type(text)))
    # elif six.PY2:
    #     if isinstance(text, str):
    #         return text.decode("utf-8", "ignore")
    #     elif isinstance(text, unicode):
    #         return text
    #     else:
    #         raise ValueError("Unsupported string type: %s" % (type(text)))
    else:
        raise ValueError("Not running on Python2 or Python 3?")


def printable_text(text):
    """Returns text encoded in a way suitable for print or `tf.logging`."""

    # These functions want `str` for both Python2 and Python3, but in one case
    # it's a Unicode string and in the other it's a byte string.
    if six.PY3:
        if isinstance(text, str):
            return text
        elif isinstance(text, bytes):
            return text.decode("utf-8", "ignore")
        else:
            raise ValueError("Unsupported string type: %s" % (type(text)))
    # elif six.PY2:
    #     if isinstance(text, str):
    #         return text
    #     elif isinstance(text, unicode):
    #         return text.encode("utf-8")
    #     else:
    #         raise ValueError("Unsupported string type: %s" % (type(text)))
    else:
        raise ValueError("Not running on Python2 or Python 3?")


def load_vocab(vocab_file):
    """Loads a vocabulary file into a dictionary."""
    vocab = collections.OrderedDict()
    index = 0
    with open(vocab_file, "r") as reader:
        # with tf.gfile.GFile(vocab_file, "r") as reader:
        while True:
            token = convert_to_unicode(reader.readline())
            if not token:
                break
            token = token.strip()
            vocab[token] = index
            index += 1
    return vocab


def convert_by_vocab(vocab, items):
    """Converts a sequence of [tokens|ids] using the vocab."""
    output = []
    for item in items:
        output.append(vocab[item])
    return output


def convert_tokens_to_ids(vocab, tokens):
    return convert_by_vocab(vocab, tokens)


def convert_ids_to_tokens(inv_vocab, ids):
    return convert_by_vocab(inv_vocab, ids)


def whitespace_tokenize(text):
    """Runs basic whitespace cleaning and splitting on a piece of text."""
    text = text.strip()
    if not text:
        return []
    tokens = text.split()
    return tokens


class FullTokenizer(object):
    """Runs end-to-end tokenziation."""

    def __init__(self, vocab_file, do_lower_case=True):
        self.vocab = load_vocab(vocab_file)
        self.inv_vocab = {v: k for k, v in self.vocab.items()}
        self.basic_tokenizer = BasicTokenizer(do_lower_case=do_lower_case)
        self.wordpiece_tokenizer = WordpieceTokenizer(vocab=self.vocab)

    def tokenize(self, text):
        split_tokens = []
        for token in self.basic_tokenizer.tokenize(text):
            for sub_token in self.wordpiece_tokenizer.tokenize(token):
                split_tokens.append(sub_token)

        return split_tokens

    def convert_tokens_to_ids(self, tokens):
        return convert_by_vocab(self.vocab, tokens)

    def convert_ids_to_tokens(self, ids):
        return convert_by_vocab(self.inv_vocab, ids)


class BasicTokenizer(object):
    """Runs basic tokenization (punctuation splitting, lower casing, etc.)."""

    def __init__(self, do_lower_case=True):
        """Constructs a BasicTokenizer.

        Args:
          do_lower_case: Whether to lower case the input.
        """
        self.do_lower_case = do_lower_case

    def tokenize(self, text):
        """Tokenizes a piece of text."""
        text = convert_to_unicode(text)
        text = self._clean_text(text)

        # This was added on November 1st, 2018 for the multilingual and Chinese
        # models. This is also applied to the English models now, but it doesn't
        # matter since the English models were not trained on any Chinese data
        # and generally don't have any Chinese data in them (there are Chinese
        # characters in the vocabulary because Wikipedia does have some Chinese
        # words in the English Wikipedia.).
        text = self._tokenize_chinese_chars(text)

        orig_tokens = whitespace_tokenize(text)
        split_tokens = []
        for token in orig_tokens:
            if self.do_lower_case:
                token = token.lower()
                token = self._run_strip_accents(token)
            split_tokens.extend(self._run_split_on_punc(token))

        output_tokens = whitespace_tokenize(" ".join(split_tokens))
        return output_tokens

    def _run_strip_accents(self, text):
        """Strips accents from a piece of text."""
        text = unicodedata.normalize("NFD", text)
        output = []
        for char in text:
            cat = unicodedata.category(char)
            if cat == "Mn":
                continue
            output.append(char)
        return "".join(output)

    def _run_split_on_punc(self, text):
        """Splits punctuation on a piece of text."""
        chars = list(text)
        i = 0
        start_new_word = True
        output = []
        while i < len(chars):
            char = chars[i]
            if _is_punctuation(char):
                output.append([char])
                start_new_word = True
            else:
                if start_new_word:
                    output.append([])
                start_new_word = False
                output[-1].append(char)
            i += 1

        return ["".join(x) for x in output]

    def _tokenize_chinese_chars(self, text):
        """Adds whitespace around any CJK character."""
        output = []
        for char in text:
            cp = ord(char)
            if self._is_chinese_char(cp):
                output.append(" ")
                output.append(char)
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _is_chinese_char(self, cp):
        """Checks whether CP is the codepoint of a CJK character."""
        # This defines a "chinese character" as anything in the CJK Unicode block:
        #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
        #
        # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
        # despite its name. The modern Korean Hangul alphabet is a different block,
        # as is Japanese Hiragana and Katakana. Those alphabets are used to write
        # space-separated words, so they are not treated specially and handled
        # like the all of the other languages.
        if ((cp >= 0x4E00 and cp <= 0x9FFF) or (
                cp >= 0x3400 and cp <= 0x4DBF) or (
                cp >= 0x20000 and cp <= 0x2A6DF) or (
                cp >= 0x2A700 and cp <= 0x2B73F) or (
                cp >= 0x2B740 and cp <= 0x2B81F) or (
                cp >= 0x2B820 and cp <= 0x2CEAF) or (
                cp >= 0xF900 and cp <= 0xFAFF) or (
                cp >= 0x2F800 and cp <= 0x2FA1F)):  #
            return True

        return False

    def _clean_text(self, text):
        """Performs invalid character removal and whitespace cleanup on text."""
        output = []
        for char in text:
            cp = ord(char)
            if cp == 0 or cp == 0xfffd or _is_control(char):
                continue
            if _is_whitespace(char):
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)


class WordpieceTokenizer(object):
    """Runs WordPiece tokenziation."""

    def __init__(self, vocab, unk_token="[UNK]", max_input_chars_per_word=200):
        self.vocab = vocab
        self.unk_token = unk_token
        self.max_input_chars_per_word = max_input_chars_per_word

    def tokenize(self, text):
        """Tokenizes a piece of text into its word pieces.

        This uses a greedy longest-match-first algorithm to perform tokenization
        using the given vocabulary.

        For example:
          input = "unaffable"
          output = ["un", "##aff", "##able"]

        Args:
          text: A single token or whitespace separated tokens. This should have
            already been passed through `BasicTokenizer.

        Returns:
          A list of wordpiece tokens.
        """

        text = convert_to_unicode(text)

        output_tokens = []
        for token in whitespace_tokenize(text):
            chars = list(token)
            if len(chars) > self.max_input_chars_per_word:
                output_tokens.append(self.unk_token)
                continue

            is_bad = False
            start = 0
            sub_tokens = []
            while start < len(chars):
                end = len(chars)
                cur_substr = None
                while start < end:
                    substr = "".join(chars[start:end])
                    if start > 0:
                        substr = "##" + substr
                    if substr in self.vocab:
                        cur_substr = substr
                        break
                    end -= 1
                if cur_substr is None:
                    is_bad = True
                    break
                sub_tokens.append(cur_substr)
                start = end

            if is_bad:
                output_tokens.append(self.unk_token)
            else:
                output_tokens.extend(sub_tokens)
        return output_tokens


def _is_whitespace(char):
    """Checks whether `chars` is a whitespace character."""
    # \t, \n, and \r are technically contorl characters but we treat them
    # as whitespace since they are generally considered as such.
    if char == " " or char == "\t" or char == "\n" or char == "\r":
        return True
    cat = unicodedata.category(char)
    if cat == "Zs":
        return True
    return False


def _is_control(char):
    """Checks whether `chars` is a control character."""
    # These are technically control characters but we count them as whitespace
    # characters.
    if char == "\t" or char == "\n" or char == "\r":
        return False
    cat = unicodedata.category(char)
    if cat.startswith("C"):
        return True
    return False


def _is_punctuation(char):
    """Checks whether `chars` is a punctuation character."""
    cp = ord(char)
    # We treat all non-letter/number ASCII as punctuation.
    # Characters such as "^", "$", and "`" are not in the Unicode
    # Punctuation class but we treat them as punctuation anyways, for
    # consistency.
    if ((cp >= 33 and cp <= 47) or (
            cp >= 58 and cp <= 64) or (
            cp >= 91 and cp <= 96) or (
            cp >= 123 and cp <= 126)):
        return True
    cat = unicodedata.category(char)
    if cat.startswith("P"):
        return True
    return False



from unittest import TestCase
from unittest.mock import patch, Mock, call

from ..expand_by_entities import (
    expand_by_entities,
    _aggregate_entities,
)
from ..elements import (
    Datum,
    Entity,
    Intent,
)


class ExpandByEntitiesTestCase(TestCase):

    def setUp(self):
        self.utterance = '我想訂明天從紐約飛到新加坡的機票'
        self.entities = [
            Entity(label=1, value='明天', start=3, end=5, replacements=['下禮拜二']),
            Entity(label=2, value='紐約', start=6, end=8),
            Entity(label=3, value='新加坡', start=10, end=13, replacements=['斯堪地那維亞', 'KIX']),
        ]
        self.intents = [
            Intent(label=0),
        ]
        self.datum = Datum(
            utterance=self.utterance,
            intents=self.intents,
            entities=self.entities,
        )

        self.expected_utterances = [
            '我想訂明天從紐約飛到新加坡的機票',
            '我想訂明天從紐約飛到斯堪地那維亞的機票',
            '我想訂明天從紐約飛到KIX的機票',
            '我想訂下禮拜二從紐約飛到新加坡的機票',
            '我想訂下禮拜二從紐約飛到斯堪地那維亞的機票',
            '我想訂下禮拜二從紐約飛到KIX的機票',
        ]
        self.expected_entity_lists = [
            [
                Entity(label=1, value='明天', start=3, end=5),
                Entity(label=2, value='紐約', start=6, end=8),
                Entity(label=3, value='新加坡', start=10, end=13),
            ],
            [
                Entity(label=1, value='明天', start=3, end=5),
                Entity(label=2, value='紐約', start=6, end=8),
                Entity(label=3, value='斯堪地那維亞', start=10, end=16),
            ],
            [
                Entity(label=1, value='明天', start=3, end=5),
                Entity(label=2, value='紐約', start=6, end=8),
                Entity(label=3, value='KIX', start=10, end=13),
            ],
            [
                Entity(label=1, value='下禮拜二', start=3, end=7),
                Entity(label=2, value='紐約', start=8, end=10),
                Entity(label=3, value='新加坡', start=12, end=15),
            ],
            [
                Entity(label=1, value='下禮拜二', start=3, end=7),
                Entity(label=2, value='紐約', start=8, end=10),
                Entity(label=3, value='斯堪地那維亞', start=12, end=18),
            ],
            [
                Entity(label=1, value='下禮拜二', start=3, end=7),
                Entity(label=2, value='紐約', start=8, end=10),
                Entity(label=3, value='KIX', start=12, end=15),
            ],
        ]
        self.expected_intents = [
            [Intent(label=0)],
            [Intent(label=0)],
            [Intent(label=0)],
            [Intent(label=0)],
            [Intent(label=0)],
            [Intent(label=0)],
        ]
        self.expected_data = []
        self.expected_data_without_replacement = []
        for utt, ents, ints in zip(
                self.expected_utterances,
                self.expected_entity_lists,
                self.expected_intents,
            ):
            self.expected_data.append(Datum(
                utterance=utt,
                entities=ents,
                intents=ints,
            ))

    def test_include_original_value(self):
        result = expand_by_entities(
            self.datum,
            include_orig=True,
        )

        # correct expand numbers
        self.assertEqual(6, len(result))
        for idx, datum in enumerate(result):
            with self.subTest(datum_idx=idx):
                self.assertIn(datum, self.expected_data)

    def test_exclude_original_value(self):
        result = expand_by_entities(
            self.datum,
            include_orig=False,
        )
        # correct expand numbers
        self.assertEqual(2, len(result))
        for idx, datum in enumerate(result):
            with self.subTest(datum_idx=idx):
                self.assertIn(datum, self.expected_data[4:])

    def test_datum_without_entities(self):
        datum_wo_entities = Datum(
            utterance='薄餡亂入',
            intents=[Intent(label=0)],
        )
        for include_orig in [True, False]:
            with self.subTest(include_orig_or_not=include_orig):
                result = expand_by_entities(datum_wo_entities, include_orig)
                self.assertEqual([datum_wo_entities], result)

    @patch('uttut.expand_by_entities.partition_by_entities', return_value=([[]], []))
    def test_expand_by_entities_call_partition_correctly(self, patch_partition):
        expand_by_entities(
            self.datum,
            include_orig=True,
        )
        patch_partition.assert_called_once_with(self.datum, True)

    @patch('uttut.expand_by_entities.partition_by_entities', return_value=([[]], []))
    def test_expand_by_entities_call_partition_correctly_false(self, patch_partition):
        expand_by_entities(
            self.datum,
            include_orig=False,
        )
        patch_partition.assert_called_once_with(self.datum, False)

    fake_parts = [['a', 'b'], ['c']]
    fake_entity_labels = [1, None]

    @patch(
        'uttut.expand_by_entities.partition_by_entities',
        return_value=(fake_parts, fake_entity_labels),
    )
    @patch('uttut.expand_by_entities.get_kth_combination', return_value=['a', 'c'])
    def test_expand_by_entities_correctly_use_sampling_method(self, patch_get_k, _):
        mock_sampling_method = Mock(return_value=[0, 1])

        expand_by_entities(
            self.datum,
            sampling_method=mock_sampling_method,
            include_orig=True,
        )

        mock_sampling_method.assert_called_once_with(2)
        patch_get_k.assert_has_calls([
            call(self.fake_parts, 0),
            call(self.fake_parts, 1),
        ])

    def test__aggregate_entities(self):
        segments = ['我想喝', '珍奶', '半糖']
        entity_names = [None, 1, 2]

        expected_entities = [
            Entity(label=1, value='珍奶', start=3, end=5),
            Entity(label=2, value='半糖', start=5, end=7),
        ]

        actual_entities = _aggregate_entities(segments, entity_names)
        self.assertEqual(expected_entities, actual_entities)



cdef class Replacement:  # noqa: E999

    cdef public unsigned int start
    cdef public unsigned int end
    cdef public object new_value
    cdef public object annotation


cdef class ReplacementGroup:   # noqa: E999

    cdef list _replacements
    cdef bint _is_done

    cpdef void add(
        self,
        unsigned int start,
        unsigned int end,
        object new_value,
        object annotation=?,
    )
    cpdef void done(self, bint skip_sort=?) except *
    cpdef bint is_empty(self)

    cdef void _validate_new_values(self) except *



import pytest

from ..add_sos_eos import AddSosEos
from ..tokens import START_TOKEN, END_TOKEN
from .common_tests import OperatorTestTemplate, ParamTuple


class TestAddSosEos(OperatorTestTemplate):

    params = [
        ParamTuple(
            ['alvin', '喜歡', '吃', '榴槤'],
            [1, 2, 3, 4],
            ['<sos>', 'alvin', '喜歡', '吃', '榴槤', '<eos>'],
            [0, 1, 2, 3, 4, 0],
            'zh',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return AddSosEos()

    def test_not_equal(self, op):
        custom_op = AddSosEos(start_token='custom_SOS', end_token='custom_EOS')
        assert custom_op != op


@pytest.mark.parametrize(
    "op, expected_configs",
    [
        pytest.param(
            AddSosEos(),
            {'start_token': START_TOKEN, 'end_token': END_TOKEN},
            id="default",
        ),
        pytest.param(
            AddSosEos('1', '2'),
            {'start_token': '1', 'end_token': '2'},
            id="no keywords",
        ),
        pytest.param(
            AddSosEos('1'),
            {'start_token': '1', 'end_token': END_TOKEN},
            id="partial input",
        ),
        pytest.param(
            AddSosEos(end_token='2', start_token='1'),
            {'start_token': '1', 'end_token': '2'},
            id="input with key",
        ),
    ],
)
def test_correct_configs(op, expected_configs):
    assert op.configs == expected_configs



import pytest

from ...tests.common_tests import OperatorTestTemplate, ParamTuple
from ..whitespace_tokenizer import WhiteSpaceTokenizer


class TestWhiteSpaceTokenizer(OperatorTestTemplate):

    params = [
        ParamTuple(
            "a \t \t \nb c",
            [1, 0, 0, 0, 0, 0, 0, 2, 0, 3],
            ["a", "b", "c"],
            [1, 2, 3],
            id='eng',
        ),
        ParamTuple(
            "  a \t \t \nb c\n\r",
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 2, 0, 3, 0, 0],
            ["a", "b", "c"],
            [1, 2, 3],
            id='eng with whitespace at head and tail',
        ),
        ParamTuple(
            "GB亂入",
            [2, 2, 2, 2],
            ["GB亂入"],
            [2],
            id='zh',
        ),
        ParamTuple(
            "",
            [],
            [],
            [],
            id='empty string',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return WhiteSpaceTokenizer()

    def test_equal(self, op):
        assert WhiteSpaceTokenizer() == op



import os
import six
import tempfile

import pytest

from .tokenization import FullTokenizer as BertFullTokenizer
from ..full import full_pipe, vocab_tokens
from uttut.elements import Datum


test_cases = [
    pytest.param(  # ref: Bert tokenization_test.py#L45
        u"UNwant\u00E9d,running",
        id='eng + accent',
    ),
]


@pytest.fixture
def tokenizer():
    with tempfile.NamedTemporaryFile(delete=False) as vocab_writer:
        if six.PY2:
            vocab_writer.write("".join([x + "\n" for x in vocab_tokens]))
        else:
            vocab_writer.write("".join(
                [x + "\n" for x in vocab_tokens]).encode("utf-8"))

        vocab_file = vocab_writer.name

    tokenizer = BertFullTokenizer(vocab_file)
    yield tokenizer
    os.unlink(vocab_file)


@pytest.mark.parametrize("input_str", test_cases)
def test_all(input_str, tokenizer):
    datum = Datum(input_str)
    output, _, _, _, _ = full_pipe.transform(datum)
    expected_output = tokenizer.tokenize(input_str)
    assert expected_output == output



from .replacement import ReplacementGroup
from .utils import (
    _transform_sequence,
    _gen_inverse_replacement_group,
)


def apply(input_str: str, replacement_group: ReplacementGroup) -> str:
    n_replacement = len(replacement_group)
    output = [''] * (2 * n_replacement + 1)
    output_lst = _transform_sequence(
        input_seq=input_str,
        replacement_group=replacement_group,
        output=output,
    )
    output_str = ''.join(output_lst)
    return output_str


def inverse(input_str: str, replacement_group: ReplacementGroup) -> ReplacementGroup:
    inverse_replacement_group = _gen_inverse_replacement_group(
        input_seq=input_str,
        replacement_group=replacement_group,
    )
    assert len(inverse_replacement_group) == len(replacement_group)
    return inverse_replacement_group



from ..span import Span


def test_normal():
    start = 10
    end = 15
    span = Span(start, end)
    assert start == span.start
    assert end == span.end


def test_all_equal():
    assert Span(10, 15) == Span(10, 15)


def test_different_type():
    assert (10, 15) != Span(10, 15)


def test_different_start():
    assert Span(10, 15) != Span(9, 15)


def test_different_end():
    assert Span(10, 15) != Span(10, 11)


def test_str():
    span = Span(10, 15)
    assert "(10, 15)" == str(span)


def test_repr():
    span = Span(10, 15)
    assert "Span(10, 15)" == repr(span)



import pytest


from ..factory import OperatorFactory


@pytest.fixture
def op_factory():
    yield OperatorFactory()


def test_register(op_factory):
    fake_op = 1
    op_factory.register('1', fake_op)
    assert op_factory['1'] == fake_op


def test_fails_to_register(op_factory):
    fake_op_1 = 1
    fake_op_2 = 2
    op_factory.register('1', fake_op_1)
    with pytest.raises(KeyError):
        op_factory.register('1', fake_op_2)


def test_operator_not_found(op_factory):
    fake_op = 1
    op_factory.register('1', fake_op)

    with pytest.raises(KeyError):
        op_factory['2']



import unicodedata

from .add_whitespace_around_character import AddWhitespaceAroundCharacter


class AddWhitespaceAroundPunctuation(AddWhitespaceAroundCharacter):

    """
    Recognize punctuation characters and add whitespace around each punctuation character

    E.g.
    >>> from uttut.pipeline.ops.add_whitespace_around_punctuation import (
        AddWhitespaceAroundPunctuation)
    >>> op = AddWhitespaceAroundPunctuation()
    >>> output_seq, label_aligner = op.transform("GB,薄餡亂入")
    >>> output_labels = label_aligner.transform([1, 1, 2, 3, 3, 4, 5])
    >>> output_seq
    "GB , 薄餡亂入"
    >>> output_labels
    [1, 1, 0, 2, 0, 3, 3, 4, 5]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 1, 2, 3, 3, 4, 5]

    """

    def _is_valid_char(self, char: str) -> bool:
        return is_punctuation(char)


def is_punctuation(char: str) -> bool:
    """Check whether char is a punctuation character.

    This code is copied from Bert `tokenization.py`.

    We treat all non-letter/number ASCII as punctuation.
    Characters such as "^", "$", and "`" are not in the Unicode
    Punctuation class but we treat them as punctuation anyways, for
    consistency.

    """
    code_point = ord(char)
    if ((
            code_point >= 33 and code_point <= 47) or (
            code_point >= 58 and code_point <= 64) or (
            code_point >= 91 and code_point <= 96) or (
            code_point >= 123 and code_point <= 126)
        ):
        return True

    cat = unicodedata.category(char)
    # For more details, please take a look at
    # https://www.fileformat.info/info/unicode/category/index.htm
    if cat.startswith("P"):
        return True

    return False



import pytest

from ..strip_accent_token import StripAccentToken
from .common_tests import OperatorTestTemplate, ParamTuple


class TestStripAccentToken(OperatorTestTemplate):

    params = [
        ParamTuple(  # ref: Bert tokenization_test.py#L64
            u"H\u00E9llo",
            [1, 2, 3, 4, 5],
            "Hello",
            [1, 2, 3, 4, 5],
            id='eng',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return StripAccentToken()

    def test_equal(self, op):
        assert op == StripAccentToken()



import pytest

from ...tests.common_tests import OperatorTestTemplate, ParamTuple
from ..eng_tokenizer import EngTokenizer


@pytest.fixture(scope='module')
def op():
    return EngTokenizer()


class TestEngTokenizer(OperatorTestTemplate):

    params = [
        ParamTuple(
            "How's it going today, Mr.Smith?",
            [1, 1, 1, 1, 1, 0, 2, 2, 0, 3, 3, 3, 3, 3, 0, 4, 4, 4, 4, 4, 5, 0,
             6, 6, 6, 6, 6, 6, 6, 6, 7],
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Mr', '.', 'Smith', '?'],
            [1, 1, 1, 2, 3, 4, 5, 6, 6, 6, 7],
            id='the same as nltk with punct',
        ),
        ParamTuple(
            "a \t \t \nb",
            [1, 0, 0, 0, 0, 0, 0, 2],
            ["a", "b"],
            [1, 2],
            id='split effect',
        ),
        ParamTuple(
            "GB亂入",
            [2, 2, 2, 2],
            ['GB亂入'],
            [2],
            id='eng + zh',
        ),
        ParamTuple(
            "",
            [],
            [],
            [],
            id='empty string',
        ),
    ]

    def op(self):
        pass

    def test_equal(self, op):
        assert EngTokenizer() == op


@pytest.mark.parametrize(
    "input_str,input_labels,expected_output_seq,"
    "expected_output_labels,expected_realigned_labels",
    [
        pytest.param(
            "  a b \n", [10, 0, 1, 3, 2, 0, 7], ['a', 'b'], [1, 2], [0, 0, 1, 0, 2, 0, 0],
            id="whitespaces_at_head_or_tail",
        ),
        pytest.param(
            "  \n", [10, 0, 1], [], [], [0, 0, 0],
            id="all whitespaces",
        ),
        pytest.param(
            "GB亂入", [2, 2, 2, 1], ['GB亂入'], [2], [2, 2, 2, 2],
            id="zh + eng",
        ),
    ],
)
def test_invertible_cases(
        input_str, input_labels,
        expected_output_seq, expected_output_labels,
        expected_realigned_labels,
        op,
    ):
    output_seq, label_aligner = op.transform(input_str)
    assert expected_output_seq == output_seq

    output_labels = label_aligner.transform(input_labels)
    assert expected_output_labels == output_labels

    realigned_labels = label_aligner.inverse_transform(expected_output_labels)
    assert expected_realigned_labels == realigned_labels



import pytest

from ...tests.common_tests import OperatorTestTemplate, ParamTuple
from ..float_token_with_space import FloatTokenWithSpace


class TestFloatTokenWithSpace(OperatorTestTemplate):

    params = [
        ParamTuple(
            "12.3 2.7 0.7777",
            [1, 1, 1, 1, 0, 2, 2, 2, 0, 3, 3, 3, 3, 3, 3],
            " _float_   _float_   _float_ ",
            [0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 0],
            id='float float float',
        ),
        ParamTuple(
            "1 2.7 1000",
            [1, 0, 2, 2, 2, 0, 3, 3, 3, 3],
            "1  _float_  1000",
            [1, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 3, 3, 3, 3],
            id='int float int',
        ),
        ParamTuple(
            "12.7.7",
            [1, 2, 3, 4, 5, 6],
            "12.7.7",
            [1, 2, 3, 4, 5, 6],
            id='invalid float',
        ),
        ParamTuple(
            "9.99",
            [0, 0, 0, 0],
            " _float_ ",
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            id='float with label 0',
        ),
        ParamTuple(
            "奇利利有12.3億元",
            [1, 1, 1, 2, 3, 3, 3, 3, 3, 3],
            "奇利利有 _float_ 億元",
            [1, 1, 1, 2, 0, 3, 3, 3, 3, 3, 3, 3, 0, 3, 3],
            id='zh with float',
        ),
        ParamTuple(
            "GB亂入",
            [1, 1, 2, 2],
            "GB亂入",
            [1, 1, 2, 2],
            id='identity',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return FloatTokenWithSpace()

    def test_equal(self, op):
        assert FloatTokenWithSpace() == op



cdef class Span:  # noqa: E999

    cdef public unsigned int start
    cdef public unsigned int end

cdef class SpanGroup:

    cdef list _spans
    cdef bint _is_done

    cpdef void add(self, unsigned int start, unsigned int end)
    cpdef void done(self) except *
    cpdef bint is_empty(self)

    cdef void _validate_contiguousness(self, list sorted_spans) except *



import pytest

from ..punctuation_except_endpoint_to_whitespace import PunctuationExceptEndpointToWhitespace
from .common_tests import OperatorTestTemplate, ParamTuple


@pytest.fixture(scope='module')
def op():
    return PunctuationExceptEndpointToWhitespace()


class TestPunctuationExceptEndpointToWhitespace(OperatorTestTemplate):

    params = [
        ParamTuple(
            "0.3",
            [1, 2, 3],
            "0.3",
            [1, 2, 3],
            id='float',
        ),
        ParamTuple(
            "O.O",
            [2, 3, 4],
            "O.O",
            [2, 3, 4],
            id='emoji',
        ),
        ParamTuple(
            "abc def",
            [2, 3, 4, 5, 6, 7, 8],
            "abc def",
            [2, 3, 4, 5, 6, 7, 8],
            id='all english',
        ),
        ParamTuple(
            "隼興亂入",
            [2, 3, 4, 5],
            "隼興亂入",
            [2, 3, 4, 5],
            id='all chinese',
        ),
    ]

    def op(self):
        pass

    def test_equal(self, op):
        assert op == PunctuationExceptEndpointToWhitespace()


@pytest.mark.parametrize(
    "input_str,input_labels,expected_output_seq,"
    "expected_output_labels,expected_realigned_labels",
    [
        pytest.param(
            "^$`", [1, 2, 3], "   ", [0, 0, 0], [0, 0, 0],
            id="puncts in the range",
        ),
        pytest.param(
            "⁕⁇※࡞", [1, 2, 3, 4], "    ", [0, 0, 0, 0], [0, 0, 0, 0],
            id="puncts in P category only",
        ),
        pytest.param(
            "西瓜汁一杯多少$$?",
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            "西瓜汁一杯多少   ",
            [1, 2, 3, 4, 5, 6, 7, 0, 0, 0],
            [1, 2, 3, 4, 5, 6, 7, 0, 0, 0],
            id="chinese + puncts",
        ),
    ],
)
def test_invertible_cases(
        input_str,
        input_labels,
        expected_output_seq,
        expected_output_labels,
        expected_realigned_labels,
        op,
    ):
    output_seq, label_aligner = op.transform(input_str)
    assert expected_output_seq == output_seq

    output_labels = label_aligner.transform(input_labels)
    assert expected_output_labels == output_labels

    realigned_labels = label_aligner.inverse_transform(expected_output_labels)
    assert expected_realigned_labels == realigned_labels



import pytest

from ..pad import Pad
from ..tokens import PAD_TOKEN
from .common_tests import OperatorTestTemplate, ParamTuple


class TestPad(OperatorTestTemplate):

    params = [
        ParamTuple(
            ['alvin', '喜歡', '吃', '榴槤', '!'],
            [1, 2, 3, 4, 5],
            ['alvin', '喜歡', '吃', '榴槤', '!'],
            [1, 2, 3, 4, 5],
            id='equal',
        ),
        ParamTuple(
            ['alvin'],
            [1],
            ['alvin', PAD_TOKEN, PAD_TOKEN, PAD_TOKEN, PAD_TOKEN],
            [1, 0, 0, 0, 0],
            id='shorter',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return Pad(5)

    def test_longer_case(self, op):
        # labels are not invertible
        output_seq, label_aligner = op.transform(
            ['alvin', '喜歡', '吃', '榴槤', '和', '蟲', '!'],
        )
        assert ['alvin', '喜歡', '吃', '榴槤', '和'] == output_seq

        output_labels = label_aligner.transform([1, 2, 3, 4, 5, 6, 7])
        assert [1, 2, 3, 4, 5] == output_labels

        output = label_aligner.inverse_transform(output_labels)
        assert [1, 2, 3, 4, 5, 0, 0] == output

    def test_equal(self, op):
        assert Pad(5) == op

    def test_not_equal(self):
        op1 = Pad(pad_token='PAD', maxlen=10)
        op2 = Pad(pad_token='PAD', maxlen=15)
        op3 = Pad(maxlen=10)
        assert op1 != op2
        assert op2 != op3


@pytest.mark.parametrize(
    "op, expected_configs",
    [
        pytest.param(
            Pad(5),
            {'pad_token': PAD_TOKEN, 'maxlen': 5},
            id="default token",
        ),
        pytest.param(
            Pad(10, 'PAD'),
            {'pad_token': 'PAD', 'maxlen': 10},
            id="input without key",
        ),
        pytest.param(
            Pad(pad_token='PAD', maxlen=1),
            {'pad_token': 'PAD', 'maxlen': 1},
            id="input with key",
        ),
    ],
)
def test_correct_configs(op, expected_configs):
    assert op.configs == expected_configs



from ..trie import Trie


def test_exist():
    word = 'abc'
    trie = Trie()
    trie.insert(word)
    assert word in trie
    assert 'ab' not in trie
    assert 'zxy' not in trie


def test_match_prefix():
    trie = Trie()
    trie.insert('珍奶')
    trie.insert('珍奶去冰')

    assert '珍奶去冰' == trie.match_prefix('珍奶去冰')
    assert '珍奶' == trie.match_prefix('珍奶去冰', shortest=True)
    assert '珍奶去冰' == trie.match_prefix('珍奶去冰謝謝')
    assert trie.match_prefix('我想要珍奶去冰') is None



import pytest

from ..token_to_index import Token2Index
from ..tokens import UNK_TOKEN
from .common_tests import OperatorTestTemplate, ParamTuple


class TestToken2Index(OperatorTestTemplate):

    params = [
        ParamTuple(
            ['薄餡', '要', '帶', '櫻花', '妹', '回來'],
            [1, 2, 3, 4, 5, 6],
            [1, 2, 3, 0, 4, 0],
            [1, 2, 3, 4, 5, 6],
            id='zh',
        ),
    ]

    @pytest.fixture
    def op(self):
        return Token2Index({UNK_TOKEN: 0, '薄餡': 1, '要': 2, '帶': 3, '妹': 4})

    @pytest.mark.parametrize(
        'error,token2index,unk_token',
        [
            pytest.param(
                ValueError, {'薄餡': 1, '要': 1, '帶': 3, '妹': 4}, UNK_TOKEN,
                id='duplicated indices',
            ),
            pytest.param(
                ValueError, {'薄餡': 0, '要': 2, '帶': 3, '妹': 4}, UNK_TOKEN,
                id='indices not continuous',
            ),
            pytest.param(
                KeyError, {'薄餡': 0, '要': 1, '帶': 2, '妹': 3}, UNK_TOKEN,
                id='token2index has no unk',
            ),
        ],
    )
    def test_invalid_token2index(self, error, token2index, unk_token):
        with pytest.raises(error):
            Token2Index(token2index=token2index, unk_token=unk_token)

    def test_equal(self, op):
        assert op == Token2Index({UNK_TOKEN: 0, '薄餡': 1, '要': 2, '帶': 3, '妹': 4})

    def test_not_equal(self):
        op1 = Token2Index(
            token2index={'薄餡': 0, '要': 1, '帶': 2, '妹': 3, 'UNK': 4},
            unk_token='UNK',
        )
        op2 = Token2Index(token2index={'薄餡': 0, '要': 1, '帶': 2, '妹': 3, UNK_TOKEN: 4})
        op3 = Token2Index(token2index={'薄餡': 0, UNK_TOKEN: 1})
        assert op1 != op2
        assert op2 != op3


@pytest.mark.parametrize(
    "op, expected_configs",
    [
        pytest.param(
            Token2Index(token2index={'薄餡': 0, '要': 1, '帶': 2, '妹': 3, UNK_TOKEN: 4}),
            {
                'token2index': {'薄餡': 0, '要': 1, '帶': 2, '妹': 3, UNK_TOKEN: 4},
                'unk_token': UNK_TOKEN,
            },
            id="default token",
        ),
        pytest.param(
            Token2Index({'薄餡': 0, '要': 1, '帶': 2, '妹': 3, '<pad>': 4}, '<pad>'),
            {
                'token2index': {'薄餡': 0, '要': 1, '帶': 2, '妹': 3, '<pad>': 4},
                'unk_token': '<pad>',
            },
            id="input without key",
        ),
        pytest.param(
            Token2Index(unk_token='PAD', token2index={'薄餡': 0, 'PAD': 1}),
            {'unk_token': 'PAD', 'token2index': {'薄餡': 0, 'PAD': 1}},
            id="input with key",
        ),
    ],
)
def test_correct_configs(op, expected_configs):
    assert op.configs == expected_configs



from ..pipe import Pipe


vocab_tokens = [
    "[UNK]", "[CLS]", "[SEP]", "want", "##want",
    "##ed", "wa", "un", "runn", "##ing",
]
vocab = {token: ind for ind, token in enumerate(vocab_tokens)}


word_piece_pipe = Pipe()
word_piece_pipe.add('WhiteSpaceTokenizer')
word_piece_pipe.add(
    'SpanSubwords', {
        'vocab': vocab,
        'unk_token': "[UNK]",
        'maxlen_per_token': 200,
    },
)



class TrieNode:

    def __init__(self):
        self._children = {}
        self._word = None

    def has_child(self, key):
        return key in self._children

    def get_child(self, key):
        return self._children[key]

    def insert_child(self, key):
        self._children[key] = TrieNode()

    @property
    def word(self):
        return self._word

    @word.setter
    def word(self, word):
        if not self._word:
            self._word = word
        else:
            raise AttributeError('Word can only be set once')


class Trie:

    def __init__(self):
        self.root = TrieNode()

    def __contains__(self, key):
        return self.exactly_search(key)

    def exactly_search(self, word: str):
        current = self.root
        for char in word:
            if not current.has_child(char):
                return False
            current = current.get_child(char)
        if current.word:
            return True
        return False

    def insert(self, word: str):
        current = self.root
        for char in word:
            if not current.has_child(char):
                current.insert_child(char)
            current = current.get_child(char)
        current.word = word

    def match_prefix(self, word: str, shortest: bool = False):
        current = self.root
        longest_word = None
        for char in word:
            if not current.has_child(char):
                return longest_word
            current = current.get_child(char)
            if current.word:
                longest_word = current.word
                if shortest:
                    return longest_word
        return longest_word



import pytest

from ..replacement import ReplacementGroup, Replacement


@pytest.mark.filterwarnings("ignore")
def test_correctly_init():
    replacement_group = ReplacementGroup()
    assert replacement_group.is_empty() is True
    assert len(replacement_group) == 0


@pytest.mark.parametrize(
    "objs",
    [
        pytest.param([(0, 0, 'a'), (0, 2, 'b')], id='str-replacement'),
        pytest.param([(0, 0, 'a', 'ohoh'), (0, 2, 'b', 'ohoh1')],
                     id='str-replacement with annotation'),
        pytest.param([(0, 0, ['a']), (0, 2, ['b'])], id='list-replacement'),
        pytest.param([(0, 0, ['a'], 'ohoh'), (0, 2, ['b'], 'ohoh1')],
                     id='list-replacement with annotation'),
    ],
)
def test_add_all(objs):
    replacement_group = ReplacementGroup.add_all(objs)
    assert len(replacement_group) == len(objs)
    assert [Replacement(*obj) for obj in objs] == replacement_group[:]


def test_add_all_empty():
    replacement_group = ReplacementGroup.add_all([])
    assert replacement_group.is_empty() is True


def test_not_done_warning():
    replacement_group = ReplacementGroup()
    with pytest.warns(UserWarning, match='ReplacementGroup needs validation, please call `done`.'):
        len(replacement_group)


def test_not_done_error():
    replacement_group = ReplacementGroup()
    replacement_group.add(1, 3, 'abc')
    with pytest.raises(RuntimeError, match='Please call `done` first.'):
        replacement_group[0]


@pytest.mark.parametrize(
    "obj,error_type",
    [
        pytest.param([(0,)], TypeError, id='lack of elements'),
        pytest.param([(0, 0, 'a', 'a', 'a')], TypeError, id='has redundant elements'),
        pytest.param([(0, 1, 'a'), (0, 2, ['b']), (3, 5, 10000)],
                     TypeError, id='mixed replacements'),
    ],
)
def test_add_fails(obj, error_type):
    with pytest.raises(error_type):
        ReplacementGroup.add_all(obj)


@pytest.mark.parametrize(
    "input1,input2",
    [
        pytest.param([], [], id='empty'),
        pytest.param([(0, 0, 'a'), (1, 2, 'b')], [(0, 0, 'a'), (1, 2, 'b')],
                     id='str-replacement'),
        pytest.param([(0, 0, ['a']), (1, 2, ['b'])], [(0, 0, ['a']), (1, 2, ['b'])],
                     id='list-replacement'),
    ],
)
def test_equal(input1, input2):
    group1 = ReplacementGroup.add_all(input1)
    group2 = ReplacementGroup.add_all(input2)
    assert group1 == group2


@pytest.mark.parametrize(
    "input1,input2",
    [
        pytest.param([(0, 0, 'a')], [(0, 0, 'a'), (1, 2, 'b')],
                     id='different length'),
        pytest.param([(0, 0, 'a'), (1, 2, 'b')], [(0, 0, 'a'), (1, 3, 'b')],
                     id='different element'),
        pytest.param([(0, 0, 'a'), (1, 2, 'b')], [(0, 0, ['a']), (1, 3, ['b'])],
                     id='different type of replacement'),
    ],
)
def test_not_equal(input1, input2):
    group1 = ReplacementGroup.add_all(input1)
    group2 = ReplacementGroup.add_all(input2)
    assert group1 != group2


def test_different_type():
    input_lst = [(0, 0, 'a')]
    assert input_lst != ReplacementGroup.add_all(input_lst)


@pytest.mark.parametrize(
    "objs,expected_objs",
    [
        pytest.param([(1, 2, 'b'), (0, 1, 'a')], [Replacement(0, 1, 'a'), Replacement(1, 2, 'b')]),
    ],
)
def test_need_sorted(objs, expected_objs):
    obj_group = ReplacementGroup.add_all(objs)
    assert expected_objs == obj_group[:]


@pytest.mark.parametrize(
    "objs",
    [
        pytest.param([(1, 10, 'b'), (2, 15, 'a')], id='intersect'),
        pytest.param([(1, 10, 'b'), (2, 8, 'a')], id='include'),
    ],
)
def test_validate_disjoint(objs):
    with pytest.raises(ValueError):
        ReplacementGroup.add_all(objs)


@pytest.mark.parametrize(
    "representation,input_lst",
    [
        pytest.param('ReplacementGroup has 1 elements', [(0, 0, 'a')], id='str-replacement'),
        pytest.param('ReplacementGroup has 0 elements', [], id='empty'),
    ],
)
def test_representation(representation, input_lst):
    group = ReplacementGroup.add_all(input_lst)
    assert representation == repr(group)



# numerical tokens
INT_TOKEN = "_int_"
FLOAT_TOKEN = "_float_"
INT_TOKEN_WITH_SPACE = f" {INT_TOKEN} "
FLOAT_TOKEN_WITH_SPACE = f" {FLOAT_TOKEN} "


#
START_TOKEN = "<sos>"
END_TOKEN = "<eos>"
UNK_TOKEN = "<unk>"
PAD_TOKEN = "<pad>"



from unittest import TestCase

from .. import ENTITY_LABEL
from ..elements import (
    Datum,
    Entity,
)
from ..exceptions import DifferentUtterance

from ..entity_overlapping_ratio import (
    check_utter_index_in_entity,
    expand_entity_to_list,
    penalty_on_same_entity_or_not,
    entity_overlapping_ratio,
)

NOT_ENTITY = ENTITY_LABEL['NOT_ENTITY']


class EntityOverlappingRatioTestCase(TestCase):

    def test_check_utter_index_in_entity(self):
        entity = Entity(label=1, value='CPH', start=0, end=2)
        ent_list = []
        utter_ind = 1
        result = check_utter_index_in_entity(utter_ind, entity, ent_list)
        self.assertEqual(result, True)

    def test_expand_entity_to_list(self):
        datum = Datum(
            utterance='家豪大大',
            intents=[],
            entities=[
                Entity(label=1, value='家豪', start=0, end=2),
            ],
        )
        ans_ent_list = [1, 1, NOT_ENTITY, NOT_ENTITY]
        test_ent_list = expand_entity_to_list(datum)
        self.assertListEqual(test_ent_list, ans_ent_list)

    def test_penalty_on_same_entity_or_not(self):
        ent1 = NOT_ENTITY
        ent2 = 1
        pen1 = penalty_on_same_entity_or_not(ent1, ent2, 5)
        self.assertEqual(pen1, 1)

        ent1 = 1
        ent2 = 2
        pen2 = penalty_on_same_entity_or_not(ent1, ent2, 5)
        self.assertEqual(pen2, 5)

        ent1 = 1
        ent2 = 1
        pen3 = penalty_on_same_entity_or_not(ent1, ent2, 5)
        self.assertEqual(pen3, 0)

    def test_raise_if_diff_utt(self):
        datum1 = Datum(
            utterance='家豪大大喜歡吃豪大大雞排',
            intents=[],
            entities=[
                Entity(label=1, value='家豪', start=0, end=2),
            ],
        )
        datum2 = Datum(
            utterance='家豪大喜歡吃豪大大雞排',
            intents=[],
            entities=[
                Entity(label=1, value='家豪', start=0, end=2),
            ],
        )
        with self.assertRaises(DifferentUtterance):
            entity_overlapping_ratio(datum1, datum2, 2)

    def test_score_1(self):
        datum1 = Datum(
            utterance='家豪大大喜歡吃豪大大雞排',
            intents=[],
            entities=[
                Entity(label=1, value='家豪', start=0, end=2),
                Entity(label=2, value='豪大', start=7, end=9),
                Entity(label=3, value='大', start=9, end=10),
                Entity(label=4, value='雞排', start=10, end=12),
            ],
        )
        datum2 = Datum(
            utterance='家豪大大喜歡吃豪大大雞排',
            intents=[],
            entities=[
                Entity(label=1, value='家豪', start=0, end=2),
                Entity(label=2, value='豪大', start=7, end=9),
                Entity(label=3, value='大', start=9, end=10),
                Entity(label=4, value='雞排', start=10, end=12),
            ],
        )

        ratio = entity_overlapping_ratio(datum1, datum2, 2)
        self.assertEqual(ratio, 1.0)

    def test_score_0(self):
        datum1 = Datum(
            utterance='吃豪大大雞排',
            intents=[],
            entities=[
                Entity(label=2, value='豪大', start=1, end=3),
                Entity(label=3, value='大', start=3, end=4),
                Entity(label=4, value='雞排', start=4, end=6),
            ],
        )
        datum2 = Datum(
            utterance='吃豪大大雞排',
            intents=[],
            entities=[
                Entity(label=2, value='豪大', start=1, end=3),
                Entity(label=5, value='大雞排', start=3, end=6),
            ],
        )

        ratio = entity_overlapping_ratio(datum1, datum2, 2)
        self.assertEqual(ratio, 0.0)

    def test_score_between_0_1(self):
        datum1 = Datum(
            utterance='家豪大大喜歡吃豪大大雞排',
            intents=[],
            entities=[
                Entity(label=1, value='家豪', start=0, end=2),
                Entity(label=6, value='大', start=9, end=10),
                Entity(label=4, value='雞排', start=10, end=12),
            ],
        )
        datum2 = Datum(
            utterance='家豪大大喜歡吃豪大大雞排',
            intents=[],
            entities=[
                Entity(label=1, value='家豪', start=0, end=2),
                Entity(label=5, value='大雞排', start=9, end=12),
            ],
        )

        ratio = entity_overlapping_ratio(datum1, datum2, 2)
        self.assertEqual(ratio, 0.5)



# distutils: sources = uttut/pipeline/ops/utils/MurmurHash3.cpp


cdef extern from "MurmurHash3.h":  # noqa: E999
    void MurmurHash3_x86_32(
        void *key,  # noqa: E225
        int len,
        unsigned int seed,
        void *out,  # noqa: E225
    )


def consistent_hash(input_str: str, mod_int: int):
    return consistent_hash_in_c(
        input_str=input_str,
        mod_int=mod_int,
        seed=0,
    )


cdef unsigned int consistent_hash_in_c(  # noqa: E999
        str input_str,
        unsigned int mod_int,
        unsigned int seed,
    ):
    cdef unsigned int out
    cdef bytes byte_str = input_str.encode('utf-8')
    MurmurHash3_x86_32(
        <char*> byte_str,  # noqa: E225, E226
        len(byte_str),
        seed,
        &out,  # noqa: E225
    )
    return out % mod_int



from typing import List

from .base import Tokenizer, TokenizerAligner


class CharTokenizerAligner(TokenizerAligner):

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        pass

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        pass


class CharTokenizer(Tokenizer):

    """Character level tokenizer

    E.g.
    >>> from uttut.pipeline.ops.char_tokenizer import CharTokenizer
    >>> op = CharTokenizer()
    >>> output_seq, label_aligner = op.transform("a b")
    >>> output_labels = label_aligner.transform([1, 2, 3])
    >>> output_seq
    ["a", "b"]
    >>> output_labels
    [1, 3]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 0, 3]

    """

    _label_aligner_class = CharTokenizerAligner

    def _tokenize(self, input_str: str) -> List[str]:
        return list(input_str)



from .elements import Datum, Entity
from .exceptions import DifferentUtterance
from . import ENTITY_LABEL

NOT_ENTITY = ENTITY_LABEL['NOT_ENTITY']


def check_utter_index_in_entity(
        utter_ind: int,
        entity: Entity,
        ent_list: list,
    ) -> bool:
    if utter_ind >= entity.start and utter_ind < entity.end:
        ent_list.append(entity.label)
    elif utter_ind >= entity.end:
        return False
    else:
        ent_list.append(NOT_ENTITY)
    return True


def expand_entity_to_list(
        datum: Datum,
    ) -> list:
    ent_list = []
    ent_ind = 0
    utter_len = len(datum.utterance)
    max_ent_ind = len(datum.entities)

    ind = -1
    while True:
        ind += 1
        if ind >= utter_len:
            break
        if ent_ind >= max_ent_ind:
            ent_list.append(NOT_ENTITY)
            continue
        cur_ent = datum.entities[ent_ind]
        if not check_utter_index_in_entity(ind, cur_ent, ent_list):
            ent_ind += 1
            ind -= 1
    return ent_list


def penalty_on_same_entity_or_not(
        ent1: int,
        ent2: int,
        wrong_penalty_rate: float,
    ) -> float:
    if ent1 == ent2:
        return 0
    else:
        if ent1 == NOT_ENTITY or ent2 == NOT_ENTITY:
            return 1
        else:
            return wrong_penalty_rate


def entity_overlapping_ratio(
        datum1: Datum,
        datum2: Datum,
        wrong_penalty_rate: float,
    ) -> float:
    if not datum1.has_same_utterance_as(datum2):
        raise DifferentUtterance("""Only two data with same utterance can be compared
got:
{}
{}
""".format(datum1.utterance, datum2.utterance))

    penalty = 0.
    ent1_list = expand_entity_to_list(datum1)
    ent2_list = expand_entity_to_list(datum2)
    for ent1, ent2 in zip(ent1_list, ent2_list):
        penalty += penalty_on_same_entity_or_not(
            ent1, ent2, wrong_penalty_rate)
    score = 1 - penalty / len(datum1.utterance)
    return score



from typing import List, Tuple

from .base import Operator, LabelAligner
from ..edit.replacement import ReplacementGroup
from ..edit import str2str
from ..edit.label_propagation import propagate_by_replacement_group
from .add_whitespace_around_punctuation import is_punctuation
from .label_transducer import get_not_entity


class PunctuationExceptEndpointToWhitespace(Operator):

    """
    Recognize punctuation characters except endpoint (`.`) and
    replace it with whitespace (" ")

    E.g.
    >>> from uttut.pipeline.ops.punctuation_except_endpoint_to_whitespace import (
        PunctuationExceptEndpointToWhitespace)
    >>> op = PunctuationExceptEndpointToWhitespace()
    >>> output_seq, label_aligner = op.transform("GB,薄餡亂入")
    >>> output_labels = label_aligner.transform([1, 1, 2, 3, 3, 4, 5])
    >>> output_seq
    "GB 薄餡亂入"
    >>> output_labels
    [1, 1, 0, 2, 3, 3, 4, 5]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 1, 0, 2, 3, 3, 4, 5]

    """

    _input_type = str
    _output_type = str

    def _transform(self, input_sequence: str) -> Tuple[str, 'LabelAligner']:
        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = str2str.apply(input_sequence, forward_replacement_group)

        label_aligner = PunctuationExceptEndpointToWhitespaceAligner(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )
        return output_sequence, label_aligner

    def _gen_forward_replacement_group(self, input_str: str) -> ReplacementGroup:
        replacement_group = ReplacementGroup()

        for i, char in enumerate(input_str):
            if is_punctuation(char) and not self._is_endpoint(char):
                replacement_group.add(
                    start=i,
                    end=i + 1,
                    new_value=" ",
                    annotation="replace-punctuations-with-whitespace-except-endpoint",
                )
        replacement_group.done()
        return replacement_group

    def _is_endpoint(self, char: str) -> bool:
        if char == ".":
            return True
        return False


class PunctuationExceptEndpointToWhitespaceAligner(LabelAligner):

    def _transform(self, labels: List[int]):
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=self._forward_edit,
            transduce_func=self._forward_transduce_func,
        )

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_not_entity(labels, output_size)

    def _inverse_transform(self, labels):
        inverse_replacement_group = str2str.inverse(self._input_sequence, self._forward_edit)
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=inverse_replacement_group,
            transduce_func=self._backward_transduce_func,
        )

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_not_entity(labels, output_size)



import pytest

from ..lowercase import Lowercase
from .common_tests import OperatorTestTemplate, ParamTuple


class TestLowercase(OperatorTestTemplate):

    params = [
        ParamTuple(
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
            list(range(26)),
            "abcdefghijklmnopqrstuvwxyz",
            list(range(26)),
            id='all a-z',
        ),
        ParamTuple(
            "Hello hoW are U",
            [1, 2, 3, 4, 5, 0, 6, 7, 8, 0, 9, 10, 11, 0, 12],
            "hello how are u",
            [1, 2, 3, 4, 5, 0, 6, 7, 8, 0, 9, 10, 11, 0, 12],
            id='mixed uppercase and lowercase eng',
        ),
        ParamTuple(
            "薄餡亂入",
            [0, 1, 2, 3],
            "薄餡亂入",
            [0, 1, 2, 3],
            id='zh',
        ),
        ParamTuple(
            "0123456789",
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "0123456789",
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            id='digits',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return Lowercase()

    def test_equal(self, op):
        assert op == Lowercase()



from typing import List

from .span import SpanGroup


def apply(input_lst: List[str], span_group: SpanGroup) -> str:

    _validate_compatibility(input_lst, span_group)

    # apply span group (list -> str)
    output_str = ''.join(input_lst)

    return output_str


def _validate_compatibility(input_lst: List[str], span_group: SpanGroup):
    if len(input_lst) != len(span_group):
        raise ValueError('Input list and span group have different length.')
    for i, (token, span) in enumerate(zip(input_lst, span_group)):
        if len(token) != span.end - span.start:
            raise ValueError(f"{i}-th element is not compatible.")



import abc
import json
import inspect
from typing import Any, List, Tuple

from .factory import OperatorFactory


op_factory = OperatorFactory()


class Operator(abc.ABC):

    """Base class for Operators

    Sub-classes should implement `_transform`

    Attributes:
        input_type: input type of sequence to transform
        output_type: output type of transformed sequence

    """

    _input_type = None
    _output_type = None

    @classmethod
    def __init_subclass__(cls):
        if cls.is_abstract():
            return

        cls.assert_has_class_attributes(['_input_type', '_output_type'])
        op_factory.register(cls.__name__, cls)

        original_init = cls.__init__

        def __init__(self, *args, **kwargs):
            original_init(self, *args, **kwargs)
            argspec = inspect.getfullargspec(original_init)
            self._configs = {
                arg_name: val
                for arg_name, val in zip(
                    argspec.args[1:],  # omit `self`
                    args,
                )
            }
            self._configs.update(kwargs)
            if argspec.defaults:
                self._configs.update({
                    arg_name: default_val
                    for arg_name, default_val in zip(
                        reversed(argspec.args),
                        reversed(argspec.defaults),
                    )
                    if arg_name not in self._configs
                })

        cls.__init__ = __init__

    @classmethod
    def is_abstract(cls):
        if getattr(cls, '__abstractmethods__', None):
            return True
        return False

    @classmethod
    def assert_has_class_attributes(cls, attrs):
        for attr_name in attrs:
            assert getattr(cls, attr_name, None) is not None, \
                f"Concrete class: {cls} should declare `{attr_name}`!"

    @classmethod
    def deserialize(cls, serialized_str: str):
        params = json.loads(serialized_str)
        return cls.from_dict(params)

    @staticmethod
    def from_dict(params):
        cls_name = params['op_name']
        kwargs = params['op_kwargs']
        return op_factory[cls_name](**kwargs)

    def serialize(self) -> str:
        return json.dumps({
            'op_name': self.__class__.__name__,
            'op_kwargs': self.configs,
        })

    @property
    def configs(self):
        return self._configs

    def __eq__(self, other):
        return type(self) == type(other) and self.configs == other.configs

    @property
    def input_type(self):
        return self._input_type

    @property
    def output_type(self):
        return self._output_type

    def transform(self, input_sequence) -> Tuple[Any, 'LabelAligner']:
        """Transform input_sequence

        Transform input_sequence to certain form which meets the output_type
        and updates labels if necessary.

        Args:
            input_sequence (input_type): utterance or tokens

        Returns:
            output_sequence (output_type): the transformed result
            label_aligner (obj): an instance of LabelAligner

        """

        self._validate_input(input_sequence)
        output_sequence, label_aligner = self._transform(input_sequence)
        self._validate_output(output_sequence)
        return output_sequence, label_aligner

    def _validate_input(self, input_sequence):
        if not isinstance(input_sequence, self.input_type):
            raise TypeError('Invalid input type')

    def _validate_output(self, output_sequence):
        if not isinstance(output_sequence, self.output_type):
            raise TypeError('Invalid output type')

    @abc.abstractmethod
    def _transform(self, input_sequence) -> Tuple[Any, 'LabelAligner']:
        pass


class LabelAligner(abc.ABC):

    """Base class for LabelAligners

    Sub-classes should implement `_transform` and `_inverse_transform`

    Attributes:
        input_sequence: utterance or tokens
        output_length (int): the length of transformed sequence
        edit: process of transformation documented by `edit` structure

    """

    def __init__(self, input_sequence, edit, output_length: int):
        self._input_length = len(input_sequence)
        self._output_length = output_length

        self._input_sequence = input_sequence
        self._forward_edit = edit

    def transform(self, labels: List[int]) -> List[int]:
        """Update labels according to forward edit

        Args:
            labels (ints): has same length as the input_sequence of Operator.transfrom

        Raise:
            ValueError if length of labels is not matched.

        Return:
            labels (ints): has same length as the output_sequence of Operator.transform

        """
        self._validate_input(labels)
        output_labels = self._transform(labels)
        self._validate_output(output_labels)
        return output_labels

    def inverse_transform(self, labels: List[int]) -> List[int]:
        """Realign model predictions to original input

        Args:
            labels (ints): has same length as the output_sequence of Operator.transfrom

        Raise:
            ValueError if length of labels is not matched.

        Return:
            labels (ints): has same length as the input_sequence of Operator.transform

        """
        self._validate_output(labels)
        output_labels = self._inverse_transform(labels)
        self._validate_input(output_labels)
        return output_labels

    def _validate_input(self, labels: List[int]):
        if len(labels) != self._input_length:
            raise ValueError('Invalid input labels')

    def _validate_output(self, labels: List[int]):
        if len(labels) != self._output_length:
            raise ValueError('Invalid output labels')

    @abc.abstractmethod
    def _transform(self, labels: List[int]) -> List[int]:
        pass

    @abc.abstractmethod
    def _inverse_transform(self, labels: List[int]) -> List[int]:
        pass


class NullLabelAligner(LabelAligner):

    def _transform(self, labels: List[int]):
        return labels

    def _inverse_transform(self, labels):
        return labels



from typing import List, Tuple

from .base import Operator, LabelAligner
from .tokens import PAD_TOKEN
from .label_transducer import get_not_entity

from ..edit import lst2lst
from ..edit.replacement import ReplacementGroup
from ..edit.label_propagation import propagate_by_replacement_group


class Pad(Operator):

    """
    Pad sequence to the given length using PAD_TOKEN (<pad>).

    E.g.
    >>> from uttut.pipeline.ops.pad import Pad
    >>> op = Pad(5)
    >>> output_seq, label_aligner = op.transform(['apple'])
    >>> output_labels = label_aligner.transform([1])
    >>> output_seq
    ['apple', '<pad>', '<pad>', '<pad>', '<pad>']
    >>> output_labels
    [1, 0, 0, 0, 0]
    >>> label_aligner.inverse_transform(output_labels)
    [1]

    """

    _input_type = list
    _output_type = list

    def __init__(self, maxlen: int, pad_token: str = PAD_TOKEN):
        self.pad_token = pad_token
        self.maxlen = maxlen

    def _transform(self, input_sequence: List[str]) -> Tuple[List[str], 'LabelAligner']:

        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = lst2lst.apply(input_sequence, forward_replacement_group)

        label_aligner = PadAligner(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )

        return output_sequence, label_aligner

    def _gen_forward_replacement_group(self, input_lst: List[str]) -> ReplacementGroup:

        seqlen = len(input_lst)
        diff = self.maxlen - seqlen

        if seqlen == self.maxlen:
            replacement_group = ReplacementGroup.add_all([])
        elif seqlen > self.maxlen:
            replacement_group = ReplacementGroup.add_all(
                [(self.maxlen, seqlen, [])],
            )
        else:
            replacement_group = ReplacementGroup.add_all(
                [(seqlen, seqlen, [self.pad_token] * diff)],
            )
        return replacement_group


class PadAligner(LabelAligner):

    def _transform(self, labels: List[int]) -> List[int]:
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=self._forward_edit,
            transduce_func=self._forward_transduce_func,
        )

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_not_entity(labels, output_size)

    def _inverse_transform(self, labels: List[int]) -> List[int]:
        inverse_replacement_group = lst2lst.inverse(self._input_sequence, self._forward_edit)
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=inverse_replacement_group,
            transduce_func=self._backward_transduce_func,
        )

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_not_entity(labels, output_size)



from typing import Dict


class Intermediate:

    """Collect intermediate products when Pipe run transformation

    attributes:
        collection (list): store input intermediates
        checkpoints (dict): (key, value) = (name, index)

    """

    def __init__(self, checkpoints: Dict[str, int]):
        self._collection: list = []
        self._checkpoints = checkpoints

    def add(self, intermediate):
        """Append intermediate into self._collection

        Arg:
            intermediate: object to be stored in self.collection

        """
        self._collection.append(intermediate)

    def get_from_checkpoint(self, name: str):
        """Retrieve intermediate in collection according to name of checkpoint

        Arg:
            name (str): name of checkpoint

        Raise:
            KeyError if name is not in self._checkpoints
            IndexError if self._checkpoints[name] >= len(self._collection)

        """
        if name not in self._checkpoints:
            raise KeyError(f"{name} is not found.")
        ck_index = self._checkpoints[name]
        return self._collection[ck_index]

    def __getitem__(self, key):
        """Retrieve intermediates by index or slice

        Arg:
            key (int or slice)

        Return:
            intermediates (list)

        Raise:
            IndexError: list index out of range if input index >= len(self.collection)

        """
        return self._collection[key]



import json
from typing import Mapping

from .. import NOT_ENTITY
from ..elements import Datum, Intent, Entity
from .base import BaseTransformer


class OrdinalLabel(BaseTransformer):
    """Transformer that turn raw label into integers"""

    def __init__(
            self,
            intent2index: Mapping[str, int],
            entity2index: Mapping[str, int],
            not_entity_index: int = 0,
        ):
        if OrdinalLabel.is_valid_mapping(intent2index):
            self._intent2index = intent2index
        else:
            raise ValueError('Intent mapping is not valid.')

        self._not_entity_index = not_entity_index
        if OrdinalLabel.is_valid_mapping({**entity2index, NOT_ENTITY: not_entity_index}):
            self._entity2index = entity2index
        else:
            raise ValueError('Entity mapping is not valid.')

    @staticmethod
    def is_valid_mapping(str2idx: Mapping[str, int]) -> bool:
        idxs = sorted(list(str2idx.values()))
        return idxs == list(range(len(idxs)))

    @property
    def not_entity_index(self):
        return self._not_entity_index

    def index2entity(self, idx):
        for entity, current_idx in self._entity2index.items():
            if idx == current_idx:
                return entity
        else:
            raise KeyError("Entity index {} not found".format(idx))

    def index2intent(self, idx):
        for intent, current_idx in self._intent2index.items():
            if idx == current_idx:
                return intent
        else:
            raise KeyError("Intent index {} not found".format(idx))

    def humanize(self, datum):
        raw_dict = {}
        raw_dict['utterance'] = datum.utterance
        raw_dict['intent'] = {
            'names': [self.index2intent(intent.label) for intent in datum.intents],
        }
        if datum.has_entities():
            entities = []

            for entity in datum.entities:
                raw_entity = {
                    'name': self.index2entity(entity.label),
                    'start': entity.start,
                    'end': entity.end,
                }
                if not entity.no_replacements():
                    raw_entity['replacements'] = list(entity.replacements)
                entities.append(raw_entity)
            raw_dict['entities'] = entities
        return raw_dict

    def machanize(self, raw_dict):
        utterance = raw_dict['utterance']
        intents = [Intent(self._intent2index[name]) for name in raw_dict['intent']['names']]
        entities = []
        if 'entities' in raw_dict:
            for raw_entity in raw_dict['entities']:
                label = self._entity2index[raw_entity['name']]
                start = raw_entity['start']
                end = raw_entity['end']
                value = utterance[start: end]
                replacements = raw_entity.get('replacements')
                entities.append(Entity(
                    label,
                    value,
                    start,
                    end,
                    replacements,
                ))
        datum = Datum(utterance, intents, entities)
        return datum

    def serialize(self) -> str:
        serializable = {
            'intent2index': self._intent2index,
            'entity2index': self._entity2index,
            'not_entity_index': self._not_entity_index,
        }
        return json.dumps(serializable)

    @classmethod
    def deserialize(cls, serialized):
        loaded = json.loads(serialized)
        intent2index = loaded['intent2index']
        entity2index = loaded['entity2index']
        not_entity_index = loaded['not_entity_index']
        return cls(intent2index, entity2index, not_entity_index)

    @classmethod
    def from_raw_dictionary(cls, raw_dict, not_entity_index=0):
        if not isinstance(not_entity_index, int):
            raise TypeError(f"not_entity_index must be integer, got {not_entity_index}"
                            f"with type{type(not_entity_index)}")
        intents = [i for d in raw_dict['data'] for i in d['intent']['names']]
        entities = [e['name'] for d in raw_dict['data'] for e in d.get('entities', [])]

        intent2index = {}
        idx = 0
        for intent in intents:
            if intent not in intent2index:
                intent2index[intent] = idx
                idx += 1
            continue

        entity2index = {}
        idx = 0
        for entity in entities:
            if entity not in entity2index:
                if idx == not_entity_index:
                    idx += 1
                entity2index[entity] = idx
                idx += 1
            continue

        return cls(intent2index, entity2index, not_entity_index=not_entity_index)



from ..intermediate import Intermediate


def _transform(input_lst, index):
    output = input_lst + [index]
    return output


def test_all():
    intm = Intermediate({'1': 1, '3': 3})
    input_lst = [0]
    expected_record = []
    for i in range(10):
        input_lst = _transform(input_lst, i)
        intm.add(input_lst)
        expected_record.append(input_lst.copy())
    assert expected_record == intm[:]
    assert expected_record[1] == intm.get_from_checkpoint('1')
    assert expected_record[3] == intm.get_from_checkpoint('3')



from typing import List, Tuple
import warnings

from .validation cimport (  # noqa: E999, E211
    _validate_start_end_in_c,
    _validate_disjoint_in_c,
)


cdef class Span:
    '''Integer offsets (start_i, end_i) of a sequence

    Args:
        start (int): start index (inclusive)
        end (int): end index (exclusive)
    '''

    def __cinit__(self, unsigned int start, unsigned int end):
        _validate_start_end_in_c(start, end)
        self.start, self.end = start, end

    def __eq__(self, other):
        if not isinstance(other, Span):
            return False
        same_start = other.start == self.start
        same_end = other.end == self.end
        return same_start and same_end

    def __str__(self):
        return f"({self.start}, {self.end})"

    def __repr__(self):
        return f"Span({self.start}, {self.end})"

    def __hash__(self):
        return hash((self.start, self.end))


cdef class SpanGroup:

    def __init__(self):
        self._spans = []
        self._is_done = False

    cpdef void add(self, unsigned int start, unsigned int end):  # type: ignore
        cdef Span span
        span = Span(start=start, end=end)
        self._spans.append(span)

    cpdef void done(self) except *:

        cdef unsigned int n_spans = len(self._spans)

        if n_spans != 0:
            self._spans = _sorted_spans(self._spans)
            _validate_disjoint_in_c(self._spans)
            self._validate_contiguousness(self._spans)
        self._is_done = True

    cdef void _validate_contiguousness(self, list sorted_spans) except *:

        cdef unsigned int n_spans, idx
        cdef Span span_current, span_next

        n_spans = len(sorted_spans)

        if sorted_spans[0].start != 0:
            raise ValueError('Spans should start from 0')

        for idx in range(n_spans - 1):
            span_current = sorted_spans[idx]
            span_next = sorted_spans[idx + 1]
            if span_current.end != span_next.start:
                raise ValueError('Spans should be contiguous.')

    @classmethod
    def add_all(cls, list spans):  # type: ignore

        cdef SpanGroup span_group
        cdef tuple span
        cdef unsigned int start, end

        span_group = cls()

        for span in spans:
            if len(span) == 2:
                start, end = span
                span_group.add(start, end)
            else:
                raise ValueError('Number of elements should = 2.')
        span_group.done()
        return span_group

    cpdef bint is_empty(self):
        return len(self._spans) == 0

    def __eq__(self, other):
        self._warn_not_done()
        if not isinstance(other, SpanGroup):
            return False
        same_length = len(other) == len(self._spans)
        same_elements = set(other) == set(self._spans)
        return same_length and same_elements

    def __getitem__(self, value):
        if not self._is_done:
            raise RuntimeError('Please call `done` first.')
        return self._spans[value]

    def __len__(self):
        self._warn_not_done()
        return len(self._spans)

    def _warn_not_done(self):
        if not self._is_done:
            warnings.warn('SpanGroup needs validation, please call `done`.')

    def __repr__(self):
        return str(self.__class__.__name__)


cdef list _sorted_spans(list spans):
    return sorted(spans, key=lambda e: e.end)  # set -> list



from .add_whitespace_around_character import AddWhitespaceAroundCharacter


class AddWhitespaceAroundCJK(AddWhitespaceAroundCharacter):

    """
    Recognize CJK characters and add whitespace around each CJK character

    E.g.
    >>> from uttut.pipeline.ops.add_whitespace_around_cjk import AddWhitespaceAroundCJK
    >>> op = AddWhitespaceAroundCJK()
    >>> output_seq, label_aligner = op.transform("GB亂入")
    >>> output_labels = label_aligner.transform([1, 1, 2, 3])
    >>> output_seq
    "GB 亂  入 "
    >>> output_labels
    [1, 1, 0, 2, 0, 0, 3, 0]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 1, 2, 3]

    """

    def _is_valid_char(self, char: str) -> bool:
        """Check whether input char is the codepoint of a CJK character.

        This code is copied from Bert `tokenization.py`.

        This defines a "chinese character" as anything in the CJK Unicode block:
        https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)

        Note that the CJK Unicode block is NOT all Japanese and Korean characters,
        despite its name. The modern Korean Hangul alphabet is a different block,
        as is Japanese Hiragana and Katakana. Those alphabets are used to write
        space - separated words, so they are not treated specially and handled
        like the all of the other languages.

        """
        code_point = ord(char)
        if ((
                code_point >= 0x4E00 and code_point <= 0x9FFF) or (
                code_point >= 0x3400 and code_point <= 0x4DBF) or (
                code_point >= 0x20000 and code_point <= 0x2A6DF) or (
                code_point >= 0x2A700 and code_point <= 0x2B73F) or (
                code_point >= 0x2B740 and code_point <= 0x2B81F) or (
                code_point >= 0x2B820 and code_point <= 0x2CEAF) or (
                code_point >= 0xF900 and code_point <= 0xFAFF) or (
                code_point >= 0x2F800 and code_point <= 0x2FA1F)
            ):
            return True

        return False



import abc


class BaseTransformer(abc.ABC):

    @abc.abstractmethod
    def humanize(self, datum):
        pass

    @abc.abstractmethod
    def machanize(self, raw_dict):
        pass

    @abc.abstractmethod
    def serialize(self):
        pass

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, serialized):
        pass



from typing import List

from uttut.elements import Datum


def get_unique_entities(
        data: List[Datum],
    )-> List:
    unique_entities = []
    for datum in data:
        if datum.entities is not None:
            for entity in datum.entities:
                unique_entities.append(entity.label)
    unique_entities = list(set(unique_entities))
    return unique_entities



from unittest.mock import Mock, call

import pytest

from uttut import NOT_ENTITY
from uttut.elements import Datum, Intent, Entity
from ..ordinal_label import OrdinalLabel


TRANSFORMATION_PAIRS = [
    (
        {
            'utterance': '你好',
            'intent': {'names': ['HI']},
        },
        Datum(utterance='你好', intents=[Intent(0)]),
    ),
    (
        {
            "utterance": "我想喝珍奶半糖",
            "intent": {
                "names": ["ORDER"],
            },
            "entities": [
                {
                    "name": "ITEM",
                    "start": 3,
                    "end": 5,
                    "replacements": ["拿鐵", "多多綠", "紅茶"],
                },
                {
                    "name": "SUGAR",
                    "start": 5,
                    "end": 7,
                    "replacements": ["無糖", "全糖"],
                },
            ],
        },
        Datum(
            utterance='我想喝珍奶半糖',
            intents=[Intent(2)],
            entities=[
                Entity(1, '珍奶', 3, 5, ["拿鐵", "多多綠", "紅茶"]),
                Entity(2, '半糖', 5, 7, ["無糖", "全糖"]),
            ],
        ),
    ),
]


@pytest.fixture(scope='function')
def tx():
    intent2index = {'HI': 0, 'BYE': 1, 'ORDER': 2}
    entity2index = {'ITEM': 1, 'SUGAR': 2}
    not_entity_index = 0
    tx = OrdinalLabel(intent2index, entity2index, not_entity_index)
    return tx


@pytest.yield_fixture
def mocked_OrdinalLabel():
    orig_checker = OrdinalLabel.is_valid_mapping
    OrdinalLabel.is_valid_mapping = Mock(return_value=True)
    yield OrdinalLabel
    OrdinalLabel.is_valid_mapping = orig_checker


def test_ordinal_label_init(mocked_OrdinalLabel):
    intent2index = {'HI': 0, 'BYE': 1, 'ORDER': 2}
    entity2index = {'DRINK': 0, 'SIZE': 1}
    not_entity_index = 2
    tx = mocked_OrdinalLabel(intent2index, entity2index, not_entity_index)
    assert tx._intent2index == intent2index
    assert tx._entity2index == entity2index
    assert tx._not_entity_index == not_entity_index

    assert tx.not_entity_index == not_entity_index

    actual_calls = mocked_OrdinalLabel.is_valid_mapping.call_args_list
    expected_calls = [
        call(intent2index),
        call({**entity2index, NOT_ENTITY: 2}),
    ]
    assert expected_calls == actual_calls


def test_ordinal_label_init_raise_if_intent2index_not_valid():
    # setup
    orig_checker = OrdinalLabel.is_valid_mapping
    OrdinalLabel.is_valid_mapping = Mock(side_effect=[False, True])

    intent2index = {}
    entity2index = {}
    with pytest.raises(ValueError):
        OrdinalLabel(intent2index, entity2index)

    # tear down
    OrdinalLabel.is_valid_mapping = orig_checker


def test_ordinal_label_init_raise_if_entity2index_not_valid():
    # setup
    orig_checker = OrdinalLabel.is_valid_mapping
    OrdinalLabel.is_valid_mapping = Mock(side_effect=[True, False])

    intent2index = {}
    entity2index = {}
    with pytest.raises(ValueError):
        OrdinalLabel(intent2index, entity2index)

    # tear down
    OrdinalLabel.is_valid_mapping = orig_checker


def test_ordinal_label_is_valid_mapping():
    correct_mapping = {'HI': 0, 'BYE': 1, 'ORDER': 2}
    duplicate_mapping = {'HI': 0, 'BYE': 1, 'ORDER': 1}
    non_ordinal_mapping = {'HI': 0, 'BYE': 1, 'ORDER': 3}
    mapping_with_negative = {'HI': -1, 'BYE': 0, 'ORDER': 1}
    assert OrdinalLabel.is_valid_mapping(correct_mapping)
    assert not OrdinalLabel.is_valid_mapping(duplicate_mapping)
    assert not OrdinalLabel.is_valid_mapping(non_ordinal_mapping)
    assert not OrdinalLabel.is_valid_mapping(mapping_with_negative)


@pytest.mark.parametrize('expected, datum', TRANSFORMATION_PAIRS)
def test_ordinal_label_humanize(expected, datum, tx):
    actual = tx.humanize(datum)
    assert set(actual.keys()) == set(expected.keys())
    assert actual['utterance'] == expected['utterance']
    assert set(actual['intent']['names']) == set(expected['intent']['names'])
    assert len(actual.get('entities', [])) == len(expected.get('entities', []))

    if len(actual.get('entities', [])) > 0:
        for actual_entity, expected_entity in zip(actual['entities'], expected['entities']):
            assert actual_entity['name'] == expected_entity['name']
            assert actual_entity['start'] == expected_entity['start']
            assert actual_entity['end'] == expected_entity['end']
            actual_replacements = set(actual_entity.get('replacements', []))
            expected_replacements = set(expected_entity.get('replacements', []))
            assert actual_replacements == expected_replacements


@pytest.mark.parametrize('raw_dict, datum', TRANSFORMATION_PAIRS)
def test_ordinal_label_machanize(raw_dict, datum, tx):
    assert tx.machanize(raw_dict) == datum


def test_ordinal_label_serialization():
    intent2index = {'HI': 0, 'BYE': 1, 'ORDER': 2}
    entity2index = {'DRINK': 1, 'SIZE': 2}
    tx = OrdinalLabel(intent2index, entity2index)

    serialized = tx.serialize()
    loaded_tx = OrdinalLabel.deserialize(serialized)

    assert tx._intent2index == loaded_tx._intent2index
    assert tx._entity2index == loaded_tx._entity2index


@pytest.fixture(scope='function')
def raw_data():
    return {
        "data": [
            {
                "utterance": "你好",
                "intent": {
                    "names": ["GREETINGS"],
                },
            },
            {
                "utterance": "我想喝珍奶半糖",
                "intent": {
                    "names": ["ORDER"],
                },
                "entities": [
                    {
                        "name": "ITEM",
                        "start": 3,
                        "end": 5,
                        "replacements": ["拿鐵", "多多綠", "紅茶"],
                    },
                    {
                        "name": "SUGAR",
                        "start": 5,
                        "end": 7,
                        "replacements": ["無糖", "全糖"],
                    },
                ],
            },
        ],
    }


def test_ordinal_from_raw_dictionary(raw_data):
    tx = OrdinalLabel.from_raw_dictionary(raw_data)
    # indexes are based on the appearance order
    assert tx._intent2index == {"GREETINGS": 0, "ORDER": 1}
    assert tx._entity2index == {"ITEM": 1, "SUGAR": 2}


def test_ordinal_from_raw_dictionary_with_custom_not_entity_index(raw_data):
    tx = OrdinalLabel.from_raw_dictionary(raw_data, not_entity_index=1)
    # indexes are based on the appearance order
    assert tx._intent2index == {"GREETINGS": 0, "ORDER": 1}
    assert tx._entity2index == {"ITEM": 0, "SUGAR": 2}


def test_ordinal_from_raw_dictionary_with_custom_not_entity_index_2(raw_data):
    tx = OrdinalLabel.from_raw_dictionary(raw_data, not_entity_index=2)
    # indexes are based on the appearance order
    assert tx._intent2index == {"GREETINGS": 0, "ORDER": 1}
    assert tx._entity2index == {"ITEM": 0, "SUGAR": 1}


@pytest.mark.parametrize('wrong_entity_index', [0., 1., 100., 0.1, '0', '0.1'])
def test_ordinal_from_raw_dictionary_raise_with_wrong_type(wrong_entity_index):
    raw_data = {'data': []}
    with pytest.raises(TypeError):
        OrdinalLabel.from_raw_dictionary(raw_data, not_entity_index=wrong_entity_index)



from typing import List, Tuple

from .base import Operator, LabelAligner
from .label_transducer import get_most_common_except_not_entity
from ..edit.replacement import ReplacementGroup
from ..edit import str2str
from ..edit.label_propagation import propagate_by_replacement_group
from uttut import ENTITY_LABEL


class AddWhitespaceAroundCharacter(Operator):

    """
    Base class for Operators which recognize the characters using function `_is_valid_char`
    and add whitespace around the matched one
    """

    _input_type = str
    _output_type = str

    def _transform(self, input_sequence: str) -> Tuple[str, 'LabelAligner']:
        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = str2str.apply(input_sequence, forward_replacement_group)

        label_aligner = AddWhitespaceAroundCharAligner(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )
        return output_sequence, label_aligner

    def _gen_forward_replacement_group(self, input_str: str) -> ReplacementGroup:
        replacement_group = ReplacementGroup()

        for i, char in enumerate(input_str):
            if self._is_valid_char(char):
                replacement_group.add(
                    start=i,
                    end=i + 1,
                    new_value=f" {char} ",
                    annotation="add-whitespace-around-any-CJK-char",
                )

        replacement_group.done()
        return replacement_group

    def _is_valid_char(self, char: str) -> bool:
        raise NotImplementedError


class AddWhitespaceAroundCharAligner(LabelAligner):

    def _transform(self, labels: List[int]) -> ReplacementGroup:
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=self._forward_edit,
            transduce_func=self._forward_transduce_func,
        )

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        assert len(labels) == 1
        return [ENTITY_LABEL['NOT_ENTITY'], labels[0], ENTITY_LABEL['NOT_ENTITY']]

    def _inverse_transform(self, labels):
        inverse_replacement_group = str2str.inverse(self._input_sequence, self._forward_edit)
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=inverse_replacement_group,
            transduce_func=self._backward_transduce_func,
        )

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)



import pytest

from ...tests.common_tests import OperatorTestTemplate, ParamTuple
from ..char_tokenizer import CharTokenizer


class TestCharTokenizer(OperatorTestTemplate):

    params = [
        ParamTuple(
            "GB亂入",
            [1, 1, 2, 2],
            ["G", "B", "亂", "入"],
            [1, 1, 2, 2],
            id='zh',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return CharTokenizer()

    def test_equal(self, op):
        assert CharTokenizer() == op



from typing import List, Tuple

from uttut.pipeline.edit import str2str, str2lst
from uttut.pipeline.edit.label_propagation import (
    propagate_by_replacement_group,
    reduce_by_span_group,
    expand_by_span_group,
)

from ..base import Operator, LabelAligner


class Tokenizer(Operator):
    """
    Base class for Operators which tokenize text.
    """

    _input_type = str
    _output_type = list
    _label_aligner_class = None

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        if cls.is_abstract():
            return
        cls.assert_has_class_attributes(['_label_aligner_class'])

    def _transform(self, input_sequence: str) -> Tuple[List[str], 'LabelAligner']:
        tokens = self._tokenize(input_sequence)

        # transform sequence
        forward_replacement_group = str2lst.gen_replacement_group(input_sequence, tokens)
        temp_str = str2str.apply(input_sequence, forward_replacement_group)
        span_group = str2lst.gen_span_group(temp_str, tokens)

        label_aligner = self._label_aligner_class(
            input_sequence=input_sequence,
            edit={
                'replacement_group': forward_replacement_group,
                'span_group': span_group,
            },
            output_length=len(tokens),
        )

        return tokens, label_aligner

    def _tokenize(self, input_str: str) -> List[str]:
        raise NotImplementedError


class TokenizerAligner(LabelAligner):

    def _transform(self, labels: List[int]) -> List[int]:
        output_labels = propagate_by_replacement_group(
            labels=labels,
            replacement_group=self._forward_edit['replacement_group'],
            transduce_func=self._forward_transduce_func,
        )
        output_labels = reduce_by_span_group(output_labels, self._forward_edit['span_group'])
        return output_labels

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        raise NotImplementedError

    def _inverse_transform(self, labels: List[int]) -> List[int]:
        inverse_replacement_group = str2str.inverse(
            self._input_sequence, self._forward_edit['replacement_group'])
        output_labels = expand_by_span_group(
            labels=labels,
            span_group=self._forward_edit['span_group'],
        )
        output_labels = propagate_by_replacement_group(
            labels=output_labels,
            replacement_group=inverse_replacement_group,
            transduce_func=self._backward_transduce_func,
        )
        return output_labels

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        raise NotImplementedError



import abc
from collections import namedtuple
from typing import List


class OperatorTestTemplate(abc.ABC):

    """Common tests for Operators

    Input pytest parametrization should have 4 input arguments, including
    input_sequence, input_labels, output_sequence, output_labels.
    """
    allowed_fixtures = ['input_sequence', 'input_labels', 'output_sequence', 'output_labels']
    ParamTuple = namedtuple('ParamTuple', allowed_fixtures + ['id'])

    @property
    @abc.abstractmethod
    def params(self) -> List[ParamTuple]:
        pass

    def pytest_generate_tests(self, metafunc):
        # intersection
        needed_args = list(set(self.allowed_fixtures) & set(metafunc.fixturenames))

        if not needed_args:
            return

        def choose_needed_param(param):
            return [getattr(param, key) for key in needed_args]

        needed_params = [
            choose_needed_param(param)
            for param in self.params
        ]
        ids = [param.id for param in self.params]
        metafunc.parametrize(needed_args, needed_params, ids=ids, scope="class")

    @abc.abstractmethod  # fixture
    def op(self):
        pass

    def test_data(self, input_sequence, input_labels, output_sequence, output_labels):
        assert len(input_sequence) == len(input_labels)
        assert len(output_sequence) == len(output_labels)

    def test_transform(self, input_sequence, output_sequence, op):
        output = op.transform(input_sequence)
        assert output_sequence == output[0]

    def test_transform_labels(self, input_sequence, input_labels, output_labels, op):
        _, label_aligner = op.transform(input_sequence)
        output = label_aligner.transform(input_labels)
        assert output_labels == output

    def test_inverse_transform_labels(self, input_sequence, input_labels, output_labels, op):
        _, label_aligner = op.transform(input_sequence)
        output = label_aligner.inverse_transform(output_labels)
        assert input_labels == output


ParamTuple = OperatorTestTemplate.ParamTuple



from typing import List, Callable, Sequence

from .replacement cimport Replacement, ReplacementGroup  # noqa: E999
from .span cimport Span, SpanGroup
from uttut import ENTITY_LABEL


def propagate_by_replacement_group(
        list labels,
        ReplacementGroup replacement_group,
        transduce_func: Callable[[List[int], int], List[int]] = None,
    ) -> List[int]:
    """Map the labels[fstart_i: fend_i] to output_labels[rstart_i: rend_i]

    Note that the length of replacement_group should be the same as
    that of backward_replacement_group.
    Map the reduced label in [forward_replacement[i].start, forward_replacement[i].end)
    to output list

    Args:
        labels (list of ints)
        replacement_group (ReplacementGroup)
        transduce_func (Callable): a function that return an integer given a list of integers.

    Return:
        labels (list of ints)
    """
    cdef unsigned int output_len, i_start, o_start, fixed_len, expand_size
    cdef list output_labels
    cdef Replacement replacement

    if transduce_func is None:
        transduce_func = _get_most_common_label

    output_len = _compute_output_length(labels, replacement_group)
    output_labels = [ENTITY_LABEL['NOT_ENTITY']] * output_len

    i_start = 0
    o_start = 0

    for replacement in replacement_group:
        # before replacement
        fixed_len = replacement.start - i_start
        output_labels[o_start: o_start + fixed_len] = labels[i_start: replacement.start]
        o_start += fixed_len

        # replacement
        expand_size = len(replacement.new_value)
        output_labels[o_start: o_start + expand_size] = transduce_func(
            labels[replacement.start: replacement.end],
            expand_size,
        )

        i_start = replacement.end
        o_start += expand_size

    # final
    output_labels[o_start:] = labels[i_start:]
    return output_labels


def _get_most_common_label(list labels, unsigned int output_size=1):
    cdef unsigned int n_labels, most_common_label
    cdef list output_labels

    n_labels = len(labels)

    if n_labels == 0:
        most_common_label = ENTITY_LABEL['NOT_ENTITY']
    else:
        most_common_label = _count_and_get_most_common(labels)

    output_labels = [most_common_label] * output_size

    return output_labels


cdef unsigned int _count_and_get_most_common(list ints):
    """
    equal to from collections import Counter
    counter = Counter(ints)
    return counter.most_common()[0][0]
    """
    cdef dict counter = {}
    cdef unsigned int i, max_key, max_value, key

    for i in ints:
        if i in counter:
            counter[i] += 1
        else:
            counter[i] = 1

    max_key = 0
    max_value = 0
    for key in counter:
        if counter[key] > max_value:
            max_key = key
            max_value = counter[key]

    return max_key


cdef unsigned int _compute_output_length(input_seq, ReplacementGroup replacement_group):
    # use cases: str -> str or list -> list
    cdef unsigned int len_iters, offset, after, before
    cdef Replacement replacement

    len_iters = len(input_seq)
    offset = 0
    for replacement in replacement_group:
        after = len(replacement.new_value)
        before = replacement.end - replacement.start
        offset += after - before

    return len_iters + offset


def reduce_by_span_group(
        list labels,
        SpanGroup span_group,
        transduce_func: Callable[[List[int], int], List[int]] = None,
    ) -> List[int]:
    # use case: string -> list

    cdef unsigned int output_len, i
    cdef list output_labels
    cdef Span span

    if transduce_func is None:
        transduce_func = _get_most_common_label

    # empty span group and empty labels pair
    if (len(span_group) == 0) and (len(labels) == 0):
        return labels

    if len(labels) != span_group[-1].end:
        raise ValueError('labels and span_group are not compatible.')

    output_len = len(span_group)
    output_labels = [ENTITY_LABEL['NOT_ENTITY']] * output_len

    for i, span in enumerate(span_group):
        output_labels[i: i + 1] = transduce_func(labels[span.start: span.end], 1)

    return output_labels


def expand_by_span_group(list labels, SpanGroup span_group) -> List[int]:
    # use case: list -> str

    cdef unsigned int output_len, label, i
    cdef list output_labels
    cdef Span span

    if len(span_group) != len(labels):
        raise ValueError('labels and span_group are not compatible.')

    # empty span group and empty labels pair
    if len(span_group) == 0:
        return labels

    output_len = span_group[-1].end
    output_labels = [ENTITY_LABEL['NOT_ENTITY']] * output_len

    for span, label in zip(span_group, labels):
        for i in range(span.start, span.end):
            output_labels[i] = label

    return output_labels



import pytest

from ..replacement import ReplacementGroup
from ..span import SpanGroup
from ..label_propagation import (
    propagate_by_replacement_group,
    reduce_by_span_group,
    expand_by_span_group,
)


test_cases = [
    pytest.param(
        [1, 1, 1, 0, 0, 3, 0, 2, 2],
        ReplacementGroup.add_all([(0, 3, 'a'), (5, 6, 'abc'), (7, 9, 'abcd')]),
        [1, 0, 0, 3, 3, 3, 0, 2, 2, 2, 2],
        ReplacementGroup.add_all([(0, 1, 'abc'), (3, 6, 'a'), (7, 11, 'ab')]),
        id='str-modify',
    ),
    pytest.param(
        [1, 1, 1, 0, 0, 3, 0, 2, 2],
        ReplacementGroup.add_all([(0, 3, ['a']), (5, 6, ['a', 'b', 'c']),
                                  (7, 9, ['a', 'b', 'c', 'd'])]),
        [1, 0, 0, 3, 3, 3, 0, 2, 2, 2, 2],
        ReplacementGroup.add_all([(0, 1, ['a', 'b', 'c']), (3, 6, ['a']),
                                  (7, 11, ['a', 'b'])]),
        id='list-modify',
    ),
    pytest.param(
        [1, 2, 3, 4],
        ReplacementGroup.add_all([(0, 0, 'a'), (1, 1, 'abc'), (4, 4, 'abcd')]),
        [0, 1, 0, 0, 0, 2, 3, 4, 0, 0, 0, 0],
        ReplacementGroup.add_all([(0, 1, ''), (2, 5, ''), (8, 12, '')]),
        id='str-insert',
    ),
    pytest.param(
        [1, 2, 3, 4],
        ReplacementGroup.add_all([(0, 0, ['a']), (1, 1, ['a', 'b', 'c']),
                                  (4, 4, ['a', 'b', 'c', 'd'])]),
        [0, 1, 0, 0, 0, 2, 3, 4, 0, 0, 0, 0],
        ReplacementGroup.add_all([(0, 1, []), (2, 5, []), (8, 12, [])]),
        id='list-insert',
    ),
]

not_invertible_test_cases = [
    pytest.param(
        [1, 2, 3, 4, 5, 6],
        ReplacementGroup.add_all([(0, 1, ''), (2, 4, ''), (5, 6, '')]),
        [2, 5],
        ReplacementGroup.add_all([(0, 0, 'a'), (1, 1, 'ab'), (2, 2, 'c')]),
        id='str-delete',
    ),
    pytest.param(
        [1, 2, 3, 4, 5, 6],
        ReplacementGroup.add_all([(0, 1, []), (2, 4, []), (5, 6, [])]),
        [2, 5],
        ReplacementGroup.add_all([(0, 0, ['a']), (1, 1, ['a', 'b']), (2, 2, ['c'])]),
        id='list-delete',
    ),
]


@pytest.mark.parametrize(
    "input_label,forward_replacements,output_label,inverse_replacements",
    test_cases + not_invertible_test_cases,
)
def test_forward_propagate_by_replacement_group(
        input_label, forward_replacements, output_label, inverse_replacements):
    output = propagate_by_replacement_group(input_label, forward_replacements)
    assert output_label == output


@pytest.mark.parametrize(
    "input_label,forward_replacements,output_label,inverse_replacements",
    test_cases,
)
def test_backward_propagate_by_replacement_group(
        input_label, forward_replacements, output_label, inverse_replacements):
    output = propagate_by_replacement_group(output_label, inverse_replacements)
    assert input_label == output


@pytest.mark.parametrize(
    "input_label,forward_replacements,output_label,inverse_replacements",
    not_invertible_test_cases,
)
def test_not_invertible(input_label, forward_replacements, output_label, inverse_replacements):
    output = propagate_by_replacement_group(output_label, inverse_replacements)
    assert [0, 2, 0, 0, 5, 0] == output


def test_propagate_by_span_group():
    span_group = SpanGroup.add_all([(0, 2), (2, 5)])
    output_label = reduce_by_span_group([2, 2, 3, 5, 5], span_group)
    assert [2, 5] == output_label
    reverse_label = expand_by_span_group(output_label, span_group)
    assert reverse_label == [2, 2, 5, 5, 5]


@pytest.mark.parametrize(
    "labels,func,spans",
    [
        pytest.param(
            [1, 2, 3, 4, 5],
            reduce_by_span_group,
            SpanGroup.add_all([(0, 2), (2, 4)]),
            id='max length',
        ),
        pytest.param(
            [1, 2, 3],
            expand_by_span_group,
            SpanGroup.add_all([(0, 2), (2, 4)]),
            id='number',
        ),
    ],
)
def test_span_group_not_competible(labels, func, spans):
    with pytest.raises(ValueError):
        func(labels, spans)


@pytest.mark.parametrize(
    "by,group",
    [
        pytest.param(
            propagate_by_replacement_group,
            ReplacementGroup.add_all([(0, 2, 'a'), (2, 5, 'b')]),
            id='replacement-group',
        ),
        pytest.param(
            reduce_by_span_group,
            SpanGroup.add_all([(0, 2), (2, 5)]),
            id='span-group',
        ),
    ],
)
def test_different_transduce_func(by, group):
    def zero_transduce(x, output_size):
        return [0] * output_size
    assert [0, 0] == by([1, 2, 3, 4, 5], group, zero_transduce)


def test_intersection_case_of_span_propagation():
    labels = [1]
    span_group = SpanGroup.add_all([(0, 1)])
    assert labels == reduce_by_span_group(labels, span_group)
    assert labels == expand_by_span_group(labels, span_group)



from .base import Operator

from .lowercase import Lowercase
from .add_whitespace_around_cjk import AddWhitespaceAroundCJK
from .add_whitespace_around_punctuation import AddWhitespaceAroundPunctuation
from .strip_accent_token import StripAccentToken
from .punctuation_except_endpoint_to_whitespace import PunctuationExceptEndpointToWhitespace

from .pattern_recognizers import (
    IntToken,
    FloatToken,
    IntTokenWithSpace,
    FloatTokenWithSpace,
    MergeWhiteSpaceCharacters,
    StripWhiteSpaceCharacters,
)
from .tokenizers import (
    CharTokenizer,
    EngTokenizer,
    ZhCharTokenizer,
    WhiteSpaceTokenizer,
    CustomWordTokenizer,
)

from .add_sos_eos import AddSosEos
from .pad import Pad
from .token_to_index import Token2Index
from .span_subwords import SpanSubwords
from .token_to_index_with_hash import Token2IndexwithHash



import pytest

from ..add_end_token import AddEndToken
from ..tokens import END_TOKEN
from .common_tests import OperatorTestTemplate, ParamTuple


class TestAddEndToken(OperatorTestTemplate):

    params = [
        ParamTuple(
            ['alvin', '喜歡', '吃', '榴槤'],
            [1, 2, 3, 4],
            ['alvin', '喜歡', '吃', '榴槤', '<eos>'],
            [1, 2, 3, 4, 0],
            'zh',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return AddEndToken()

    def test_not_equal(self, op):
        custom_op = AddEndToken(end_token='custom_EOS')
        assert custom_op != op


@pytest.mark.parametrize(
    "op, expected_configs",
    [
        pytest.param(
            AddEndToken(),
            {'end_token': END_TOKEN},
            id="default",
        ),
        pytest.param(
            AddEndToken('1'),
            {'end_token': '1'},
            id="no keywords",
        ),

    ],
)
def test_correct_configs(op, expected_configs):
    assert op.configs == expected_configs


@pytest.mark.parametrize(
    "end_token, error",
    [
        pytest.param(
            None,
            ValueError,
            id="The end token can not be None",
        ),
        pytest.param(
            1,
            TypeError,
            id="wrong type of end token",
        ),
        pytest.param(
            '',
            ValueError,
            id="empty string is not allowed.",
        ),
    ],
)
def test_invalid_end_token(end_token, error):
    with pytest.raises(error):
        AddEndToken(end_token)



from typing import List, Iterable


def get_kth_combination(
        iterables: List[Iterable],
        k: int,
    ):
    return get_kth_combination_in_c(
        iterables=iterables,
        k=k,
    )


cdef list get_kth_combination_in_c(  # noqa: E999
        list iterables,
        int k,
    ):
    cdef list output
    cdef int i, n_iters, chosen_id, k_in_suffix

    n_iters = len(iterables)
    output = []
    k_in_suffix = k

    for i in range(n_iters):
        iter_ = iterables[i]
        chosen_id = k_in_suffix % len(iter_)
        output.append(iter_[chosen_id])
        k_in_suffix = k_in_suffix // len(iter_)
    return output



import re

from ..tokens import FLOAT_TOKEN
from .int_token import IntTokenAligner
from .base import PatternRecognizer


class FloatToken(PatternRecognizer):

    """
    Recognize float (ex: 12.3, 1.7) in the input string
    and replace them with FLOAT_TOKEN (_float_)

    E.g.
    >>> from uttut.pipeline.ops.float_token import FloatToken
    >>> op = FloatToken()
    >>> output_seq, label_aligner = op.transform("10.7")
    >>> output_labels = label_aligner.transform([1, 1, 1, 1])
    >>> output_seq
    "_float_"
    >>> output_labels
    [1, 1, 1, 1, 1, 1, 1]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 1, 1, 1]

    """

    _label_aligner_class = IntTokenAligner

    REGEX_PATTERN = re.compile(r"(?<![\.\d])\d+\.\d+(?![\.\d])")
    TOKEN = FLOAT_TOKEN

    def _gen_forward_replacement_group(self, input_str: str):  # type: ignore
        return super()._gen_forward_replacement_group(
            input_str=input_str,
            annotation='float-token',
        )



from typing import List, Tuple, Pattern

from uttut.pipeline.edit import str2str
from uttut.pipeline.edit.label_propagation import propagate_by_replacement_group
from uttut.pipeline.edit.replacement import ReplacementGroup
from ..base import Operator, LabelAligner


class PatternRecognizer(Operator):
    """
    Base class for Operators which recognize text patterns using regular expressions
    and replace them with a predefined token.
    """

    _input_type = str
    _output_type = str
    _label_aligner_class = None

    REGEX_PATTERN: Pattern
    TOKEN: str

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        if cls.is_abstract():
            return
        cls.assert_has_class_attributes(['_label_aligner_class'])

    def _transform(self, input_sequence: str) -> Tuple[str, 'LabelAligner']:
        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = str2str.apply(input_sequence, forward_replacement_group)

        label_aligner = self._label_aligner_class(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )
        return output_sequence, label_aligner

    def _gen_forward_replacement_group(
            self,
            input_str: str,
            annotation: str = None,
        ) -> ReplacementGroup:
        shift = 0
        replacement_group = ReplacementGroup()

        while shift < len(input_str):
            match = self.REGEX_PATTERN.search(input_str, shift)
            if match is None:
                break
            replacement_group.add(
                start=match.start(),
                end=match.end(),
                new_value=self.TOKEN,
                annotation=annotation,
            )
            shift = match.end()
        replacement_group.done()
        return replacement_group


class PatternRecognizerAligner(LabelAligner):

    def _transform(self, labels: List[int]):
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=self._forward_edit,
            transduce_func=self._forward_transduce_func,
        )

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        raise NotImplementedError

    def _inverse_transform(self, labels):
        inverse_replacement_group = str2str.inverse(self._input_sequence, self._forward_edit)
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=inverse_replacement_group,
            transduce_func=self._backward_transduce_func,
        )

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        raise NotImplementedError



from typing import List, Tuple, Dict

from .tokens import UNK_TOKEN
from .base import Operator, LabelAligner
from .label_transducer import get_most_common, get_most_common_except_not_entity
from ..edit import lst2lst
from ..edit.replacement import ReplacementGroup
from ..edit.label_propagation import propagate_by_replacement_group


class SpanSubwords(Operator):

    _input_type = list
    _output_type = list

    def __init__(
            self,
            vocab: Dict[str, int],
            unk_token: str = UNK_TOKEN,
            maxlen_per_token: int = 200,
        ):
        """
        Span each element in input_sequence to several subwords according to the input vocabulary.

        If the element or part of element is not found in vocabulary,
        the unk_token would be returned.

        # Requirement: All elements in input_sequence should not contain
                        whitespaces and accent tokens.

        This class is almost the same as WordPieceTokenizer in BERT.
        ref: https://github.com/google-research/bert/blob/master/tokenization.py

        E.g.
        >>> from uttut.pipeline.ops.span_subwords import SpanSubwords
        >>> op = SpanSubwords(vocab={"I": 0, "apple": 1, "##s": 2}, unk='<unk>', maxlen_per_token=5)
        >>> output_seq, output_labels, realigner = op.transform(["I", "like", "apples"])
        >>> output_labels = label_aligner.transform([1, 2, 3])
        >>> output_seq
        ["I", "<unk>", "apple", "##s"]
        >>> output_labels
        [1, 2, 3, 3]
        >>> label_aligner.inverse_transform(output_labels)
        [1, 2, 3]

        """

        self.vocab = vocab
        self.unk_token = unk_token
        self.maxlen_per_token = maxlen_per_token

    def _transform(self, input_sequence: List[str]) -> Tuple[List[str], 'LabelAligner']:
        """
        # Requirement: All elements in input_sequence should not contain
                        whitespaces and accent tokens.
        """
        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = lst2lst.apply(input_sequence, forward_replacement_group)

        label_aligner = SpanSubwordsAligner(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )
        return output_sequence, label_aligner

    def _gen_forward_replacement_group(self, input_lst: List[str]) -> ReplacementGroup:
        replacement_group = ReplacementGroup()
        for i, token in enumerate(input_lst):
            subtokens = span_subwords(
                word=token,
                unk_token=self.unk_token,
                vocab=self.vocab,
                max_input_chars_per_word=self.maxlen_per_token,
            )
            if subtokens != [token]:
                replacement_group.add(
                    start=i,
                    end=i + 1,
                    new_value=subtokens,
                    annotation='span-subwords',
                )
        replacement_group.done()

        return replacement_group


class SpanSubwordsAligner(LabelAligner):

    def _transform(self, labels: List[int]) -> List[int]:
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=self._forward_edit,
            transduce_func=self._forward_transduce_func,
        )

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common(labels, output_size)

    def _inverse_transform(self, labels: List[int]) -> List[int]:
        inverse_replacement_group = lst2lst.inverse(self._input_sequence, self._forward_edit)
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=inverse_replacement_group,
            transduce_func=self._backward_transduce_func,
        )

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)


def span_subwords(
        word: str,
        unk_token: str,
        vocab: Dict[str, int],
        max_input_chars_per_word: int,
    ) -> List[str]:
    """
    Tokenize a word into several subwords.

    This code is copied from Bert tokenization.py.

    Requirement: The input word should not contain whitespaces and accent tokens.

    This uses a greedy longest-match-first algorithm to perform tokenization
    using the given vocabulary.

    The default arguments in BERT are
    unk_token = '[UNK]'
    max_input_chars_per_word = 200

    E.g.
    >>> span_subwords(
            word="unaffable",
            vocab={"un": 0, "##aff": 1, "##able": 2},
            max_input_chars_per_word=200,
        )
    ["un", "##aff", "##able"]

    """

    # length of word is more than expected -> return [unk_token]
    if len(word) > max_input_chars_per_word:
        return [unk_token]

    # subword matching
    chars = list(word)
    is_bad = False
    start = 0
    sub_tokens = []
    while start < len(chars):
        end = len(chars)
        cur_substr = None
        while start < end:
            substr = "".join(chars[start: end])
            if start > 0:
                substr = "##" + substr
            if substr in vocab:
                cur_substr = substr
                break
            end -= 1
        if cur_substr is None:
            is_bad = True
            break
        sub_tokens.append(cur_substr)
        start = end

    if is_bad:
        # one of subwords is not in vocab
        return [unk_token]
    return sub_tokens



import pytest

from ..validation import _validate_start_end


def test_normal():
    _validate_start_end(1, 3)


def test_end_smaller_than_start():
    with pytest.raises(ValueError):
        _validate_start_end(18, 10)



from ..pipe import Pipe


basic_pipe = Pipe()
basic_pipe.add('Lowercase')
basic_pipe.add('AddWhitespaceAroundCJK')
basic_pipe.add('AddWhitespaceAroundPunctuation')
basic_pipe.add('MergeWhiteSpaceCharacters')
basic_pipe.add('StripWhiteSpaceCharacters')
basic_pipe.add('StripAccentToken')
basic_pipe.add('WhiteSpaceTokenizer')



import pytest

from ..add_whitespace_around_punctuation import AddWhitespaceAroundPunctuation
from .common_tests import OperatorTestTemplate, ParamTuple


class TestAddWhitespaceAroundPunctuation(OperatorTestTemplate):

    params = [
        ParamTuple(
            "^$`",
            [1, 2, 3],
            " ^  $  ` ",
            [0, 1, 0, 0, 2, 0, 0, 3, 0],
            id='punct in range',
        ),
        ParamTuple(
            "⁕⁇※࡞",  # These are just a small portion of punctuations in P category.
            [1, 2, 3, 4],
            " ⁕  ⁇  ※  ࡞ ",
            [0, 1, 0, 0, 2, 0, 0, 3, 0, 0, 4, 0],
            id='puncts in P category only',
        ),
        ParamTuple(
            "a,b",
            [1, 2, 3],
            "a , b",
            [1, 0, 2, 0, 3],
            id='eng',
        ),
        ParamTuple(
            "隼興亂入",
            [2, 3, 4, 5],
            "隼興亂入",
            [2, 3, 4, 5],
            id='all chinese',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return AddWhitespaceAroundPunctuation()

    def test_equal(self, op):
        assert op == AddWhitespaceAroundPunctuation()



from typing import List

import re

from ..label_transducer import get_not_entity
from .base import PatternRecognizer, PatternRecognizerAligner


class StripWhiteSpaceCharactersAligner(PatternRecognizerAligner):

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_not_entity(labels=labels, output_size=output_size)

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_not_entity(labels=labels, output_size=output_size)


class StripWhiteSpaceCharacters(PatternRecognizer):

    """
    Recognize leading and trailing whitespace characters in the string
    and replace them with an empty string ("")

    E.g.
    >>> from uttut.pipeline.ops.strip_whitespace_characters import StripWhiteSpaceCharacters
    >>> op = StripWhiteSpaceCharacters()
    >>> output_seq, label_aligner = op.transform(" a\n")
    >>> output_labels = label_aligner.transform([1, 2, 3])
    >>> output_seq
    "a"
    >>> output_labels
    [2]
    >>> label_aligner.inverse_transform(output_labels)
    [0, 2, 0]

    """

    _label_aligner_class = StripWhiteSpaceCharactersAligner

    REGEX_PATTERN = re.compile(r"\A\s+|\s+\Z")
    TOKEN = ""

    def _gen_forward_replacement_group(self, input_str: str):  # type: ignore
        return super()._gen_forward_replacement_group(
            input_str=input_str,
            annotation='strip whitespace',
        )



from ..pipe import Pipe


vocab_tokens = [
    "[UNK]", "[CLS]", "[SEP]", "want", "##want", "##ed", "wa", "un", "runn",
    "##ing", ",",
]
vocab = {token: ind for ind, token in enumerate(vocab_tokens)}


full_pipe = Pipe()
full_pipe.add('Lowercase')
full_pipe.add('AddWhitespaceAroundCJK')
full_pipe.add('AddWhitespaceAroundPunctuation')
full_pipe.add('MergeWhiteSpaceCharacters')
full_pipe.add('StripWhiteSpaceCharacters')
full_pipe.add('StripAccentToken')
full_pipe.add('WhiteSpaceTokenizer')
full_pipe.add(
    'SpanSubwords', {
        'vocab': vocab,
        'unk_token': "[UNK]",
        'maxlen_per_token': 200,
    },
)



from typing import List, Tuple, Dict

from .base import Operator, NullLabelAligner
from .tokens import UNK_TOKEN

from ..edit import lst2lst
from ..edit.replacement import ReplacementGroup


def validate_continuity(token2index: Dict[str, int]):
    indices = set([index for _, index in token2index.items()])
    if len(token2index) != len(indices):
        raise ValueError("duplicated index")

    start_from_zero = min(indices) == 0
    end_at_size = max(indices) == len(token2index) - 1
    if not (start_from_zero and end_at_size):
        raise ValueError("indices are not continuous.")


class Token2Index(Operator):

    """
    Map token (str) to index (int) given token2index dictionary

    Note that token2index should have UNK_TOKEN (<unk>).

    E.g.
    >>> from uttut.pipeline.ops.token_to_index import Token2Index
    >>> op = Token2Index({'unk': 0, 'I': 1, 'like': 2, 'apples': 3}, 'unk')
    >>> output_seq, label_aligner = op.transform(['I', 'like', 'apples', 'oh'])
    >>> output_labels = label_aligner.transform([3, 4, 5, 6])
    >>> output_seq
    [1, 2, 3, 0]
    >>> output_labels
    [3, 4, 5, 6]
    >>> label_aligner.inverse_transform(output_labels)
    [3, 4, 5, 6]

    """

    _input_type = list
    _output_type = list

    def __init__(self, token2index: Dict[str, int], unk_token: str = UNK_TOKEN):
        self._validate_token2index(token2index, unk_token)
        self.token2index = token2index
        self.unk_token = unk_token

    def _validate_token2index(self, token2index: Dict[str, int], unk_token: str):
        validate_continuity(token2index)
        if unk_token not in token2index:
            raise KeyError(f"token2index should have token {unk_token}")

    def _transform(self, input_sequence: List[str]) -> Tuple[List[int], 'LabelAligner']:
        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = lst2lst.apply(input_sequence, forward_replacement_group)

        label_aligner = Token2IndexAligner(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )
        return output_sequence, label_aligner

    def _gen_forward_replacement_group(self, input_lst: List[str]) -> ReplacementGroup:

        replacement_group = ReplacementGroup()
        for idx, token in enumerate(input_lst):
            if token in self.token2index:
                index = self.token2index[token]
            else:
                index = self.token2index[self.unk_token]
            replacement_group.add(
                start=idx,
                end=idx + 1,
                new_value=[index],
                annotation='token2index',
            )
        replacement_group.done()
        return replacement_group


Token2IndexAligner = NullLabelAligner



import pytest

from ...tests.common_tests import OperatorTestTemplate, ParamTuple
from ..merge_whitespace_characters import MergeWhiteSpaceCharacters


class TestMergeWhiteSpaceCharacters(OperatorTestTemplate):

    params = [
        ParamTuple(
            " \t\n\r\x0b\x0c ",
            [0, 0, 0, 0, 0, 0, 0],
            " ",
            [0],
            id='all whitespace characters',
        ),
        ParamTuple(
            " _int_   _int_   _int_ ",
            [0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0],
            " _int_ _int_ _int_ ",
            [0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0],
            id='result of int token with space op',
        ),
        ParamTuple(
            "GB亂入  _int_  次",
            [1, 1, 2, 2, 0, 0, 3, 3, 3, 3, 3, 0, 0, 4],
            "GB亂入 _int_ 次",
            [1, 1, 2, 2, 0, 3, 3, 3, 3, 3, 0, 4],
            id='zh + int token',
        ),
        ParamTuple(
            "我想要喝多 多 綠",
            [0, 0, 0, 0, 1, 1, 1, 1, 1],
            "我想要喝多 多 綠",
            [0, 0, 0, 0, 1, 1, 1, 1, 1],
            id='label identity',
        ),
        ParamTuple(
            "GB亂入",
            [1, 1, 2, 2],
            "GB亂入",
            [1, 1, 2, 2],
            id='identity',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return MergeWhiteSpaceCharacters()

    def test_equal(self, op):
        assert MergeWhiteSpaceCharacters() == op



from typing import List

from collections import Counter

from uttut import ENTITY_LABEL


def get_most_common(labels: List[int], output_size: int) -> List[int]:
    counter = Counter(labels)
    common_label = counter.most_common()[0][0]
    return [common_label] * output_size


def get_most_common_except_not_entity(labels: List[int], output_size: int) -> List[int]:
    meaningful_labels = [label for label in labels if label != ENTITY_LABEL['NOT_ENTITY']]
    if len(meaningful_labels) == 0:
        return get_not_entity(labels, output_size)
    return get_most_common(meaningful_labels, output_size)


def get_not_entity(labels: List[int], output_size: int) -> List[int]:
    return [ENTITY_LABEL['NOT_ENTITY']] * output_size



def entity_position_correct(utterance: str, entity) -> bool:
    start = entity.start
    end = entity.end
    if entity.value == utterance[start: end]:
        return True
    return False


def overlap(entity, next_entity) -> bool:
    return bool(entity.end > next_entity.start)


def msg_entity_wrong_position(utterance: str, entity) -> str:
    start = entity.start
    end = entity.end
    return """
        Utterance position {start} - {end} expected to be {expected},
        but got {result}.
    """.format(
        start=start,
        end=end,
        expected=entity.value,
        result=utterance[start: end],
    )


def msg_entity_overlapping(utterance: str, entity, next_entity) -> str:
    return """
        For utterance {utt}, entity {sv} ends at {se}, and entity {nv} starts at {ns},
        which are overlapping. Currently, uttut doesn't support overlapping or hierachy entities.
    """.format(
        utt=utterance,
        sv=entity.value,
        se=entity.end,
        nv=next_entity.value,
        ns=next_entity.start,
    )



from ..consistent_hash import consistent_hash


def test_consistent_hash():
    output = consistent_hash("隼興", 100)
    assert output == 15
    output = consistent_hash("希臘優格", 100)
    assert output == 51
    output = consistent_hash("隼興", 100)
    assert output == 15



from typing import List

import unicodedata

from ..label_transducer import get_most_common_except_not_entity
from .base import Tokenizer, TokenizerAligner


class EngTokenizerAligner(TokenizerAligner):

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)


class EngTokenizer(Tokenizer):

    """English Word level tokenizer

    All of following functions are copied from BERT.
    https://github.com/google-research/bert/blob/master/tokenization.py

    E.g.
    >>> from uttut.pipeline.ops.eng_tokenizer import EngTokenizer
    >>> op = EngTokenizer()
    >>> output_seq, label_aligner = op.transform("a b")
    >>> output_labels = label_aligner.transform([1, 2, 3])
    >>> output_seq
    ["a", "b"]
    >>> output_labels
    [1, 3]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 0, 3]

    """

    _label_aligner_class = EngTokenizerAligner

    def _tokenize(self, input_str: str) -> List[str]:
        orig_tokens = whitespace_tokenize(input_str)
        split_tokens = []
        for token in orig_tokens:
            split_tokens.extend(self._run_split_on_punc(token))
        output_tokens = whitespace_tokenize(" ".join(split_tokens))
        return output_tokens

    def _run_split_on_punc(self, text: str) -> List[str]:
        """
        Recognize punctuations and seperate them into independent tokens.

        E.g.
        1. "abc, cdf" -> ["abc", ",", " ", "cdf"]
        2. "I like apples." -> ["I", "like", "apples", "."]

        """
        chars = list(text)
        i = 0
        start_new_word = True
        output = []
        while i < len(chars):
            char = chars[i]
            if _is_punctuation(char):
                output.append([char])
                start_new_word = True
            else:
                if start_new_word:
                    output.append([])
                start_new_word = False
                output[-1].append(char)
            i += 1
        return ["".join(x) for x in output]


def whitespace_tokenize(text: str) -> List[str]:
    """Split text into list by whitespace characters

    E.g.
    1. "a  \t\t\n\n b" -> ["a", "b"]
    2. "  \n\t\n  " -> []
    3. "  a b \t\n" -> ["a", "b"]

    """
    tokens = text.split()
    return tokens


def _is_punctuation(char):
    """Checks whether `chars` is a punctuation character.

    We treat all non-letter/number ASCII as punctuation.
    Characters such as "^", "$", and "`" are not in the Unicode
    Punctuation class but we treat them as punctuation anyways, for
    consistency.

    """
    cp = ord(char)
    if (
        (cp >= 33 and cp <= 47) or (
            cp >= 58 and cp <= 64) or (
            cp >= 91 and cp <= 96) or (
            cp >= 123 and cp <= 126)
    ):
        return True
    cat = unicodedata.category(char)
    if cat.startswith("P"):
        return True
    return False



import pytest

from ..replacement import ReplacementGroup
from uttut.pipeline.edit import lst2lst


test_cases = [
    pytest.param(
        list('我想要喝200元的珍奶10杯'),
        ReplacementGroup.add_all([(4, 7, list('_int_')), (11, 13, list('_int_'))]),
        list('我想要喝_int_元的珍奶_int_杯'),
        ReplacementGroup.add_all([(4, 9, list('200')), (13, 18, list('10'))]),
        id='modify',
    ),
    pytest.param(
        list('我想喝珍奶'),
        ReplacementGroup.add_all([(0, 0, list('其實')), (0, 1, ['妳']), (1, 1, ['不']),
                                  (3, 5, list('多多綠')), (5, 5, list('對吧！'))]),
        list('其實妳不想喝多多綠對吧！'),
        ReplacementGroup.add_all([(0, 2, []), (2, 3, ['我']), (3, 4, []),
                                  (6, 9, list('珍奶')), (9, 12, [])]),
        id='insert',
    ),
    pytest.param(
        list('我不想喝珍奶'),
        ReplacementGroup.add_all([(1, 2, [])]),
        list('我想喝珍奶'),
        ReplacementGroup.add_all([(1, 1, ['不'])]),
        id='delete',
    ),
    pytest.param(
        ['這個', '句子', '沒有', '改變'],
        ReplacementGroup.add_all([]),
        ['這個', '句子', '沒有', '改變'],
        ReplacementGroup.add_all([]),
        id='identity',
    ),
]


@pytest.mark.parametrize(
    "input_lst,forward_replacements,output_lst,inverse_replacements",
    test_cases,
)
def test_forward_apply(input_lst, forward_replacements, output_lst, inverse_replacements):
    output = lst2lst.apply(input_lst, forward_replacements)
    assert output_lst == output


@pytest.mark.parametrize(
    "input_lst,forward_replacements,output_lst,inverse_replacements",
    test_cases,
)
def test_backward_apply(input_lst, forward_replacements, output_lst, inverse_replacements):
    output = lst2lst.apply(output_lst, inverse_replacements)
    assert input_lst == output


@pytest.mark.parametrize(
    "input_lst,forward_replacements,output_lst,inverse_replacements",
    test_cases,
)
def test_forward_inverse(input_lst, forward_replacements, output_lst, inverse_replacements):
    output = lst2lst.inverse(input_lst, forward_replacements)
    assert inverse_replacements == output


@pytest.mark.parametrize(
    "input_lst,forward_replacements,output_lst,inverse_replacements",
    test_cases,
)
def test_backward_inverse(input_lst, forward_replacements, output_lst, inverse_replacements):
    output = lst2lst.inverse(output_lst, inverse_replacements)
    assert forward_replacements == output



import pytest

from ...tests.common_tests import OperatorTestTemplate, ParamTuple
from ..strip_whitespace_characters import StripWhiteSpaceCharacters


class TestStripWhiteSpaceCharacters(OperatorTestTemplate):

    params = [
        ParamTuple(
            " \t\n\r\x0b\x0c ",
            [0, 0, 0, 0, 0, 0, 0],
            "",
            [],
            id='all whitespace characters',
        ),
        ParamTuple(
            " abc \n\t ",
            [0, 1, 1, 1, 0, 0, 0, 0],
            "abc",
            [1, 1, 1],
            id='eng with whitespaces',
        ),
        ParamTuple(
            "我想要喝多 多 綠",
            [0, 0, 0, 0, 1, 1, 1, 1, 1],
            "我想要喝多 多 綠",
            [0, 0, 0, 0, 1, 1, 1, 1, 1],
            id='inner whitespace',
        ),
        ParamTuple(
            "GB亂入",
            [1, 1, 2, 2],
            "GB亂入",
            [1, 1, 2, 2],
            id='identity',
        ),
        ParamTuple(
            "",
            [],
            "",
            [],
            id='empty string',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return StripWhiteSpaceCharacters()

    def test_all_whitespaces(self, op):
        # labels are not invertible
        output_seq, label_aligner = op.transform("  \n\n\t\t")
        assert "" == output_seq

        output_labels = label_aligner.transform([1, 2, 3, 4, 5, 6])
        assert [] == output_labels

        output = label_aligner.inverse_transform([])
        assert output == [0, 0, 0, 0, 0, 0]

    def test_equal(self, op):
        assert StripWhiteSpaceCharacters() == op



from .char_tokenizer import CharTokenizer
from .eng_tokenizer import EngTokenizer
from .zh_char_tokenizer import ZhCharTokenizer
from .whitespace_tokenizer import WhiteSpaceTokenizer
from .custom_word_tokenizer import CustomWordTokenizer
